from qaravan.tensorQ.statevector_sim import StatevectorSim, partial_overlap, all_zero_sv, op_action
from qaravan.core.utils import RunContext 
from qaravan.core.circuits import two_local_circ
from tqdm import tqdm
import numpy as np

def sv_environment(circ, left_sv, gate_idx): 
    mat, indices = circ[gate_idx].matrix, circ[gate_idx].indices

    circ1 = circ[:gate_idx]
    circ2 = circ[gate_idx+1:]

    sim1 = StatevectorSim(circ1, init_state=None)
    sim2 = StatevectorSim(circ2.dag(), init_state=left_sv)

    sv1 = sim1.run(progress_bar=None).reshape(2**circ.num_sites)
    sv2 = sim2.run(progress_bar=None).reshape(2**circ.num_sites)

    return partial_overlap(sv1, sv2, skip=indices), mat

def environment_update(circ, gate_idx, pre_state, post_state, target_sv, init_sv, direction='decreasing'): 
    indices = circ[gate_idx].indices
    env = partial_overlap(pre_state, post_state, skip=indices)
    x,s,yh = np.linalg.svd(env)
    new_mat = yh.conj().T @ x.conj().T
    circ.update_gate(gate_idx, new_mat)

    if direction == 'decreasing': 
        if gate_idx == 1: 
            pre_state = init_sv 
        else:
            pre_state = op_action(circ[gate_idx-1].matrix.conj().T, circ[gate_idx-1].indices, pre_state)   # undo the next gate        
        post_state = op_action(new_mat.conj().T, indices, post_state)

    else: 
        pre_state = op_action(new_mat, indices, pre_state)        
        if gate_idx == len(circ)-2: 
            post_state = target_sv
        else:
            post_state = op_action(circ[gate_idx+1].matrix, circ[gate_idx+1].indices, post_state)   # undo the next gate

    return 1 - np.abs(np.trace(new_mat @ env)), pre_state, post_state

def environment_state_prep(target_sv, init_sv=None, circ=None, skeleton=None, context=None, quiet=True):
    """ uses environment tensors to optimize a circuit to prepare target_sv 
    either circ, skeleton or context.resume_state must be provided """
    context = RunContext() if context is None else context
    
    if context.resume: 
        circ = context.run_state['circ']
        cost_list = context.run_state['cost_list']
        pre_state = context.run_state['pre_state']
        post_state = context.run_state['post_state']
    else:
        if circ is None:
            circ = two_local_circ(skeleton)
        
        sim = StatevectorSim(circ, init_state=init_sv)
        ansatz = sim.run(progress_bar=False).reshape(2**circ.num_sites)
        cost_list = [1-np.abs(target_sv.conj().T @ ansatz)]

        pre_state = op_action(circ[-1].matrix.conj().T, circ[-1].indices, ansatz)   # undo the last gate 
        post_state = target_sv

    init_sv = all_zero_sv(circ.num_sites, dense=True) if init_sv is None else init_sv
    for step in range(context.step, context.max_iter):
        left_sweep = tqdm(reversed(range(1,len(circ))), disable=quiet, desc='Left sweep')
        right_sweep = tqdm(range(len(circ)-1), disable=quiet, desc='Right sweep')

        for i in left_sweep: 
            cost, pre_state, post_state = environment_update(circ, i, pre_state, post_state, target_sv, init_sv, direction='decreasing')
            cost_list.append(cost)

        for i in right_sweep:
            cost, pre_state, post_state = environment_update(circ, i, pre_state, post_state, target_sv, init_sv, direction='increasing')
            cost_list.append(cost)
        
        run_state = {'step': step, 'circ': circ, 'cost_list': cost_list, 'pre_state': pre_state, 'post_state': post_state}
        if context.step_update(run_state):
            break

    return circ, cost_list


import numpy as np
from qaravan.core import BaseSim, pauli_I, pauli_X, pauli_Y, pauli_Z
from functools import reduce
from scipy.linalg import block_diag
import re

def is_special_orthogonal(mat): 
    return bool(np.allclose(np.dot(mat.T, mat), np.eye(mat.shape[0])) and np.allclose(np.dot(mat, mat.T), np.eye(mat.shape[0])) 
            and np.isclose(np.linalg.det(mat), 1.0))

def is_special_unitary(mat): 
    return bool(np.allclose(np.dot(mat.T.conj(), mat), np.eye(mat.shape[0])) and np.allclose(np.dot(mat, mat.T.conj()), np.eye(mat.shape[0])) 
            and np.isclose(np.linalg.det(mat), 1.0))

def random_su2():
    x = np.random.randn(4)
    x = x / np.linalg.norm(x)
    a = x[0] + 1j*x[1]
    b = x[2] + 1j*x[3]
    return np.array([[a, -b.conj()],
                     [b,  a.conj()]], dtype=complex)

def matchgate_from_blocks(A,B): 
    G = np.zeros((4, 4), dtype=complex)
    G[0, 0] = A[0,0]
    G[0, 3] = A[0,1]
    G[3, 0] = A[1,0]
    G[3, 3] = A[1,1]
    G[1, 1] = B[0,0]
    G[1, 2] = B[0,1]
    G[2, 1] = B[1,0]
    G[2, 2] = B[1,1]
    return G

def random_matchgate():
    A = random_su2()
    B = random_su2()
    return matchgate_from_blocks(A, B)

def generate_majorana_ops(n):
    majorana_ops = []
    for i in range(n): 
        kron_list = [pauli_Z] * i + [pauli_X] + [pauli_I] * (n - i - 1)
        op = reduce(np.kron, kron_list)
        majorana_ops.append(op)

        kron_list = [pauli_Z] * i + [pauli_Y] + [pauli_I] * (n - i - 1)
        op = reduce(np.kron, kron_list)
        majorana_ops.append(op)

    return majorana_ops

def generate_R(n, U, indices): 
    majorana_ops = generate_majorana_ops(2)
    m_indices = np.arange(2*indices[0], 2*indices[1]+2)

    R = np.eye(2*n, dtype=complex)
    for i, ml in zip(m_indices, majorana_ops):
        for j, mr in zip(m_indices, majorana_ops):
            R[i,j] = np.trace(mr @ U @ ml @ U.conj().T)/(2**2)

    Rp = np.array([R.T[2*i,:] + 1j*R.T[2*i+1,:] for i in range(n)])
    assert is_special_orthogonal(R), "R is not special orthogonal"
    return R, Rp

def circuit_to_R(circ): 
    R = np.eye(2*circ.num_sites, dtype=complex)
    for gate in circ.gate_list:
        if gate.name == "SWAP":
            mats = [pauli_I, pauli_X, pauli_Y, pauli_Z]
            ch = np.random.choice(np.arange(4))
            mat = mats[ch]
            matrix = np.kron(mat, mat)
        elif gate.name == "M":
            matrix = gate.matrix 
        else: 
            raise ValueError(f"Unsupported gate: {gate.name}")
        
        r, _ = generate_R(circ.num_sites, matrix, gate.indices)
        R = R @ r

    Rp = np.array([R.T[2*i,:] + 1j*R.T[2*i+1,:] for i in range(circ.num_sites)])
    return R, Rp

def generate_labels(instring, outstring):
    in_labels = []
    for i, bit in enumerate(instring):
        if bit == '1':  
            in_labels.append(f'gamma_{2*i}')

    out_labels = []
    for i, bit in enumerate(outstring):
        if bit == '0':  
            out_labels.append(f'gamma_m{i}')
            out_labels.append(f'gamma_n{i}')
        elif bit == '1':
            out_labels.append(f'gamma_n{i}')
            out_labels.append(f'gamma_m{i}')

    return in_labels + out_labels + in_labels[::-1]
 
def parse_label(label):
    m = re.match(r"gamma_([nm]?)(\d+)", label)
    if not m:
        raise ValueError(f"Label '{label}' does not match expected pattern.")
    kind = m.group(1)
    if kind == "":
        kind = "gamma"  # bare gamma, use its number directly
    idx = int(m.group(2))
    return kind, idx

def get_entry(Rp, H, label_i, label_j):
    kind_i, idx_i = parse_label(label_i)
    kind_j, idx_j = parse_label(label_j)

    if kind_i == 'm' and kind_j == 'm':
        return (Rp @ H @ Rp.T)[idx_i, idx_j]  # gamma_mi, gamma_mj
    
    elif kind_i == 'm' and kind_j == 'n':
        return (Rp @ H @ Rp.conj().T)[idx_i, idx_j] # gamma_mi, gamma_nj
    
    elif kind_i == 'm' and kind_j == 'gamma': 
        return (Rp @ H)[idx_i, idx_j]
    
    elif kind_i == 'n' and kind_j == 'm':
        return (Rp.conj() @ H @ Rp.T)[idx_i, idx_j]
    
    elif kind_i == 'n' and kind_j == 'n':
        return (Rp.conj() @ H @ Rp.conj().T)[idx_i, idx_j]
    
    elif kind_i == 'n' and kind_j == 'gamma':
        return (Rp.conj() @ H)[idx_i, idx_j]
    
    elif kind_i == 'gamma' and kind_j == 'm':
        return (H @ Rp.T)[idx_i, idx_j]
    
    elif kind_i == 'gamma' and kind_j == 'n':
        return (H @ Rp.conj().T)[idx_i, idx_j]
    
    elif kind_i == 'gamma' and kind_j == 'gamma':
        return int(idx_i == idx_j)
    
    else:
        raise ValueError(f"Invalid labels: {label_i}, {label_j}")
    
def exp_mat(n, labels, Rp): 
    H = block_diag(*[np.array([[1,1j], [-1j,1]])] * n)
    mat = np.zeros((len(labels), len(labels)), dtype=complex)
    for i, label_i in enumerate(labels):
        for j, label_j in enumerate(labels):
            if i < j:
                mat[i,j] = get_entry(Rp, H, label_i, label_j)
            elif i > j:
                mat[i,j] = -mat[j,i]

    return mat

def probability(instring, outstring, circ): 
    n = len(instring) 
    num_measured_qubits = len([bit for bit in outstring if bit != 'i'])

    labels = generate_labels(instring, outstring)
    _, Rp = circuit_to_R(circ)
    mat = exp_mat(n, labels, Rp)

    norm = 2**(2*num_measured_qubits)
    return  (np.sqrt(np.linalg.det(mat))/norm).real

def sample(instring, circ, epsilon=0.0): 
    s = ''
    prev_probs = [1.0]
    for i in range(circ.num_sites):
        outstring = s + '0' + 'i'*(circ.num_sites - i - 1)   
        p = probability(instring, outstring, circ) + np.random.normal(0, epsilon)
        if np.random.rand() < p/prev_probs[-1]: 
            s += '0'  
            normalizer = p
        else: 
            s += '1'
            normalizer = prev_probs[-1] - p
        prev_probs.append(normalizer)
        
    return s 

class MatchgateSimH(BaseSim):
    def __init__(self, circ, init_state=None):
        super().__init__(circ, init_state=init_state, nm=None)

    def initialize_state(self):
        """ internal state is a SO(2n) matrix, where n is the number of sites """
        self.state = np.eye(2*self.num_sites, dtype=complex)
        self.instring = '0' * self.num_sites if self.init_state is None else self.init_state

    def generate_R(self, gate): 
        U, indices = gate.matrix, gate.indices
        majorana_ops = generate_majorana_ops(2)
        m_indices = np.arange(2*indices[0], 2*indices[1]+2)

        R = np.eye(2*self.num_sites, dtype=complex)
        for i, ml in zip(m_indices, majorana_ops):
            for j, mr in zip(m_indices, majorana_ops):
                R[i,j] = np.trace(mr @ U @ ml @ U.conj().T)/(2**2)

        assert is_special_orthogonal(R), "R is not special orthogonal"
        return R

    def apply_gate(self, gate):
        R = self.generate_R(gate)
        self.state = R @ self.state

    def probability(self, outstring): 
        n = self.num_sites
        num_measured_qubits = len([bit for bit in outstring if bit != 'i'])

        labels = generate_labels(self.instring, outstring)
        R = self.state
        Rp = np.array([R.T[2*i,:] + 1j*R.T[2*i+1,:] for i in range(self.num_sites)])
        mat = exp_mat(n, labels, Rp)

        norm = 2**(2*num_measured_qubits)
        return  (np.sqrt(np.linalg.det(mat))/norm).real
    
    def sample(self): 
        s = ''
        prev_probs = [1.0]
        for i in range(self.num_sites):
            outstring = s + '0' + 'i'*(self.num_sites - i - 1)
            p = self.probability(outstring)
            if np.random.rand() < p/prev_probs[-1]: 
                s += '0'  
                normalizer = p
            else: 
                s += '1'
                normalizer = prev_probs[-1] - p
            prev_probs.append(normalizer)
            
        return s 

class MatchgateSimS(BaseSim):
    def __init__(self, circ, init_state=None):
        super().__init__(circ, init_state=init_state, nm=None)  

    def initialize_state(self):
        """ 
        internal state is a covariance matrix, which is a 2n x 2n matrix where n is the number of sites
        init_state can be provided as a bitstring or a covariance matrix
        """
        if self.init_state is None: 
            summands = [np.array([[0,-1], [1,0]]) for i in range(self.num_sites)]
            self.state = block_diag(*summands)

        elif type(self.init_state) == str:
            mat = np.array([[0,-1], [1,0]])
            # summand will have mat or -mat on the diagonal depending on the bitstring
            summands = [-mat if b == '1' else mat for b in self.init_state]
            self.state = block_diag(*summands)

        elif type(self.init_state) == np.ndarray:
            self.state = self.init_state 

        else:
            raise ValueError("init_state must be either a covariance matrix or a bitstring.")
        
    def apply_gate(self, gate):
        gammas = [np.kron(pauli_X, pauli_I), np.kron(pauli_Y, pauli_I), np.kron(pauli_Z, pauli_X), 
                  np.kron(pauli_Z, pauli_Y)]
        rot_matrix = np.array([[0.5 * np.trace(gamma1 @ gate.matrix.conj().T @ gamma2 @ gate.matrix) 
                               for gamma2 in gammas] for gamma1 in gammas])
        summands = [np.eye(2) for i in range(self.num_sites)]
        summands[gate.indices[0]] = rot_matrix
        summands.remove(gate.indices[1])
        full_rot_matrix = block_diag(*summands)
        self.state = full_rot_matrix @ self.state @ full_rot_matrix.conj().T


from qaravan.core.base_sim import BaseSim
from qaravan.core.utils import string_to_sv, pretty_print_sv
from qaravan.core.gates import proj_up, proj_down
from qaravan.core.circuits import Circuit
import numpy as np
import torch 
from ncon_torch import ncon, permute
from scipy.sparse import csc_matrix
import copy
from functools import reduce

class StatevectorSim(BaseSim):
    def __init__(self, circ, init_state=None, backend="numpy", device="cpu"):
        super().__init__(circ, init_state=init_state, nm=None)    
        self.backend = backend
        self.device = torch.device(device)

    def to_backend(self, array):
        if self.backend == "torch":
            if torch.is_tensor(array):
                return array.to(dtype=torch.complex128, device=self.device)
            else:
                return torch.tensor(array, dtype=torch.complex128, device=self.device)
        else:
            if isinstance(array, np.ndarray):
                return array
            else:
                return array.detach().cpu().numpy()
            
    def initialize_state(self):
        """ 
        internal state is a rank-n tensor with local dimension inherited from the circuit 
        init_state can be provided either as a tensor, a statevector or a bitstring
        """
        shape = [self.local_dim] * self.num_sites
        if self.init_state is None:
            sv = np.zeros(self.local_dim**self.num_sites, dtype=np.complex128)
            sv[0] = 1.0
        elif isinstance(self.init_state, str):
            sv = string_to_sv(self.init_state, self.circ.local_dim)
        elif isinstance(self.init_state, (np.ndarray, torch.Tensor)):
            sv = self.init_state
        else:
            raise ValueError("init_state must be a NumPy array, Torch tensor, or a bitstring")

        sv = sv.reshape(shape)
        self.state = self.to_backend(sv)
        
    def apply_gate(self, gate):
        mat = self.to_backend(gate.matrix)
        self.state = op_action(mat, gate.indices, self.state, local_dim=self.local_dim)

    def measure(self, meas_sites):
        """ returns a measurement bitstring """
        return measure_sv(self.get_statevector(), meas_sites)
        
    def measure_and_collapse(self, meas_sites): 
        """ returns a measurement bitstring and post-measurement statevector """
        return measure_and_collapse_sv(self.get_statevector(), meas_sites)
        
    def local_expectation(self, local_ops):
        """ op is a list of 1-local Hermitian matrices """
        self.run(progress_bar=False)
        right_state = copy.deepcopy(self.state)
        for i in range(self.num_sites):
            op = self.to_backend(local_ops[i])
            state_indices = [-(j+1) for j in range(self.num_sites)] 
            state_indices[i] = 1
            right_state = ncon((op, right_state), ([-(i+1),1], state_indices))

        return ncon((self.state.conj(), right_state), ([i for i in range(1, self.num_sites+1)], [i for i in range(1, self.num_sites+1)])).real

    def __str__(self):
        sv = self.state.reshape(self.local_dim**self.num_sites)
        return pretty_print_sv(sv, self.local_dim)

    def get_statevector(self): 
        if not self.ran:
            self.run(progress_bar=False)
        return self.state.reshape(self.local_dim**self.num_sites)

def locs_to_indices(locs, n): 
    shifted_locs = [loc + 1 for loc in locs]
    gate_indices = [-i for i in shifted_locs] + shifted_locs

    boundaries = [0] + shifted_locs
    tensor_indices = []
    for i in range(len(shifted_locs)):
        tensor_indices += [-j for j in range(boundaries[i] + 1, boundaries[i + 1])]
        tensor_indices.append(shifted_locs[i])
    
    tensor_indices += [-j for j in range(boundaries[-1] + 1, n+1)]
    return gate_indices, tensor_indices

def op_action(op, indices, sv, local_dim=2): 
    if op.ndim != 2*len(indices): 
        op = op.reshape(*[local_dim]*2*len(indices))
    
    n = sv.ndim if sv.ndim > 1 else int(np.log(len(sv)) / np.log(local_dim))
    state = sv.reshape(*[local_dim]*n) if sv.ndim == 1 else sv
    #state = copy.deepcopy(sv).reshape(*[local_dim]*n) if sv.ndim == 1 else copy.deepcopy(sv)
    
    # locs_to_indices assumes ascending order for indices, so sort them first and transpose the operator accordingly
    sorted_indices = sorted(indices)
    sort_order = [indices.index(i) for i in sorted_indices]
    perm = sort_order + [i + len(indices) for i in sort_order]  
    op = permute(op, perm)

    gate_indices, state_indices = locs_to_indices(sorted_indices, n)
    new_sv = ncon((op, state), (gate_indices, state_indices))
    return new_sv.reshape(local_dim**n) if sv.ndim == 1 else new_sv

def all_zero_sv(num_sites, local_dim=2, dense=False, backend="numpy"):    
    if dense: 
        sv = np.zeros(local_dim**num_sites)
        sv[0] = 1.0
    else: 
        sv = csc_matrix(([1], ([0], [0])), shape=(local_dim**num_sites, 1))
    return sv

def random_sv(num_sites, local_dim=2):
    sv = np.random.rand(local_dim**num_sites) + 1j*np.random.rand(local_dim**num_sites)
    sv /= np.linalg.norm(sv)
    return sv

def partial_overlap(sv1, sv2, local_dim=2, skip=None): 
    system_size = int(np.log(len(sv1)) / np.log(local_dim))
    sites = sorted(skip) if skip is not None else []
    
    psi = sv1.reshape([local_dim] * system_size)
    phi_conj = sv2.reshape([local_dim] * system_size).conj()

    psi_labels = [0] * system_size
    phi_conj_labels = [0] * system_size

    next_contract_label = 1
    next_free_label = -1

    for i in range(system_size):
        if i in sites:
            psi_labels[i] = next_free_label
            phi_conj_labels[i] = next_free_label - len(sites)
            next_free_label -= 1
        else:
            psi_labels[i] = next_contract_label
            phi_conj_labels[i] = next_contract_label
            next_contract_label += 1
    
    contraction = ncon([psi, phi_conj], [psi_labels, phi_conj_labels])
    kept_dim = local_dim ** len(sites)
    return contraction.reshape((kept_dim, kept_dim))   

def rdm_from_sv(sv, sites, local_dim=2):
    return partial_overlap(sv, sv, local_dim=local_dim, skip=sites) 

def measure_sv(sv, meas_sites): 
    rdm = rdm_from_sv(sv, meas_sites)
    probs = np.real(np.diag(rdm))
    basis_outcomes = [np.binary_repr(i, width=len(meas_sites)) for i in range(len(probs))]
    return str(np.random.choice(basis_outcomes, p=probs))

def measure_and_collapse_sv(sv, meas_sites, local_dim=2): 
    outcome_str = measure_sv(sv, meas_sites)
    
    bras = [np.array([1,0]) if bit == '0' else np.array([0,1]) for bit in outcome_str]
    bra = reduce(np.kron, bras).reshape([local_dim]* len(bras))
    
    n = sv.ndim if sv.ndim > 1 else int(np.log(len(sv)) / np.log(local_dim))
    state = sv.reshape(*[local_dim]*n) if sv.ndim == 1 else sv
    
    sorted_indices = sorted(meas_sites)
    sort_order = [meas_sites.index(i) for i in sorted_indices]
    perm = sort_order 

    bra = permute(bra, perm)
    gate_indices, state_indices = locs_to_indices(sorted_indices, n)
    bra_indices = gate_indices[len(gate_indices)//2:]
    collapsed = ncon((bra, state), (bra_indices, state_indices))
    collapsed /= np.linalg.norm(collapsed)
    
    collapsed = collapsed.reshape(local_dim**(n - len(meas_sites))) if sv.ndim == 1 else collapsed
    return outcome_str, collapsed


from qaravan.core.utils import *
from qaravan.core.gates import CNOT, CNOT3, embed_operator, Gate
from qaravan.core.param_gates import U, U01, RZ
from qaravan.core.paulis import pauli_Z
from qaravan.core.circuits import Circuit, circ_to_mat

from scipy.linalg import expm
from itertools import product, combinations
from scipy.optimize import minimize
from tqdm import tqdm
from numpy.linalg import qr 

# only for qubits

def dressed_cnot_circ(skeleton, num_qubits, params, local_dim=2):
    single_site = U if local_dim == 2 else U01
    two_site = CNOT if local_dim == 2 else CNOT3 
    
    counter = 0
    gate_list = []
    for pair in skeleton: 
        gate_list.append(single_site(pair[0], *params[3*counter:3*(counter+1)]))
        counter += 1

        gate_list.append(single_site(pair[1], *params[3*counter:3*(counter+1)]))
        counter += 1

        gate_list.append(two_site(pair, num_qubits))

    for i in range(num_qubits):
        gate_list.append(single_site(i, *params[3*counter:3*(counter+1)]))
        counter += 1

    return Circuit(gate_list, num_qubits)

def dressed_cnot_skeletons(num_sites, num_cnots, orientation=False, a2a=False):
    if not a2a: 
        pairs = [(i, i+1) for i in range(num_sites-1)]
    else: 
        pairs = [(i,j) for i in range(num_sites) for j in range(num_sites) if i != j]

    if orientation: 
        pairs += [pair[::-1] for pair in pairs]

    return list(product(pairs, repeat=num_cnots))

def msp_cost(params, skeleton, vecs, target): 
    num_qubits = int(np.log2(len(vecs[0])))
    ansatz_mat = circ_to_mat(dressed_cnot_circ(skeleton, num_qubits, params))
    
    c = 0.0 
    
    if type(target) == np.ndarray:
        for vec in vecs: 
            c += np.linalg.norm(ansatz_mat @ vec - target @ vec)
    else:
        for vec,out in zip(vecs, target): 
            c += np.linalg.norm(ansatz_mat @ vec - out)
        
    return c / (len(vecs)) 

import os 
def msp_optimize(vecs, target, skeleton, num_reps=10, quiet=True): 
    """ 
    target can either be the matrix or the list of output statevectors 
    vecs are the list of input statevectors
    """

    num_cnots = len(skeleton)
    num_qubits = int(np.log2(len(vecs[0])))
    num_params = (num_cnots * 6) + (num_qubits * 3)
    
    results = []
    vals = []
    for i in tqdm(range(num_reps)): 
        params = np.random.rand(num_params)
        result = minimize(msp_cost, x0=params, args=(skeleton, vecs, target), 
                                 method='BFGS')
        results.append(result)
        vals.append(result.fun)
        if not quiet:
            print(result.fun)

    print(f"Best cost with {num_cnots} CNOTs is: {min(vals)}")
    return results, vals

def brickwall_skeleton(n, num_layers): 
    skeleton = []
    for _ in range(num_layers):
        for i in range(n//2): 
            skeleton.append((2*i,2*i+1))
        for i in range((n+1)//2 -1): 
            skeleton.append((2*i+1,2*i+2))
    return skeleton 

def environment(in_state, ansatz, out_state, gate_idx): 
    d = ansatz.local_dim
    env = np.zeros((d,d), dtype=complex)
    for i in range(d): 
        for j in range(d): 
            a_copy = copy.deepcopy(ansatz)
            mat = np.zeros((d,d))
            mat[i,j] = 1.0
            a_copy.update_gate(gate_idx, mat)
            env[i,j] = out_state.conj().T @ a_copy.to_matrix() @ in_state 
            
    return env.T 

def ssp_opt(in_state, ansatz, out_state, num_sweeps=100): 
    cost_list = []
    term_cond = None
    for sweep in range(num_sweeps):
        for gate_idx in range(len(ansatz.gate_list)): 
            if ansatz.gate_list[gate_idx].name[0] == 'U': 
                gate = ansatz.gate_list[gate_idx]
                env = environment(in_state, ansatz, out_state, gate_idx) 

                x,s,yh = np.linalg.svd(env) 
                new_gate = yh.conj().T @ x.conj().T 
                ansatz.update_gate(gate_idx, new_gate)

                new_cost = 1-np.trace(env@new_gate).real
                if new_cost < 0: 
                    if np.abs(new_cost) < 1e-6: 
                        new_cost = 1e-16 
                    else: 
                        print("Cost has become negative somehow")
                
        cost_list.append(np.sqrt(new_cost))
        if sweep > 1 and (cost_list[-2] - cost_list[-1]) / cost_list[-2] < 1e-3:
            term_cond = "cost function has plateaued"
            break
        if cost_list[-1] < 1e-6:
            term_cond = "cost function has converged"
            break
            
        
    print(term_cond)
    return ansatz, cost_list, term_cond 

def msp_opt(in_states, ansatz, out_states, num_sweeps=100, quiet=True, threshold=2e-8): 
    term_cond = None
    cost_list = []
    for sweep in tqdm(range(num_sweeps)):
        for gate_idx in range(len(ansatz.gate_list)): 
            if ansatz.gate_list[gate_idx].name[0] == 'U': 
                gate = ansatz.gate_list[gate_idx]
                
                env_list = []
                for in_state,out_state in zip(in_states, out_states):
                    env = environment(in_state, ansatz, out_state, gate_idx) 
                    env_list.append(env)
                    
                env = sum(env_list)

                x,s,yh = np.linalg.svd(env) 
                new_gate = yh.conj().T @ x.conj().T 
                ansatz.update_gate(gate_idx, new_gate)

                new_cost = len(in_states)-np.trace(env@new_gate).real
                if new_cost < 0: 
                    if np.abs(new_cost) < 1e-6: 
                        new_cost = 1e-16 
                    else: 
                        print("Cost has become negative somehow")
        
        cost_list.append(np.sqrt(new_cost))
        if sweep > 1 and (cost_list[-2] - cost_list[-1]) / cost_list[-2] < 1e-4:
            term_cond = "cost function has plateaued"
            break
        if cost_list[-1] < threshold:
            term_cond = "cost function has converged"
            break
        if not quiet: 
            if sweep % 10 == 0: 
                print(cost_list[-1])
    
    term_cond = "reached maximum number of sweeps" if term_cond is None else term_cond
    print(term_cond)
    return ansatz, cost_list, term_cond

def msp_job(in_states, out_states, skeleton, local_dim=2, max_sweeps=500, max_instances=50, threshold=2e-8): 
    num_cnots = len(skeleton)
    num_sites = int(np.ceil(np.log(len(in_states[0]))/np.log(local_dim)))
    num_params = (num_cnots * 6) + (num_sites * 3)
    
    sols = []
    data = []
    for i in range(max_instances):
        params = np.random.rand(num_params)
        ansatz = dressed_cnot_circ(skeleton, num_sites, params, local_dim) 
        new_ansatz, cost_list, term_cond = msp_opt(in_states, ansatz, out_states, num_sweeps=max_sweeps, threshold=threshold)
        sols.append(new_ansatz)
        data.append({'sol':new_ansatz, 'cost_list':cost_list, 'term_cond': term_cond})

        print(f"After optimization: {cost_list[-1]}")
        if cost_list[-1] < threshold: 
            break
            
    return sols, data

def construct_unitary(in_strings, out_states, local_dim=2):
    in_mat = np.array([string_to_sv(in_str, local_dim) for in_str in in_strings])
    out_mat = np.array(out_states)
    A = out_mat.T @ in_mat
    Q, _ = qr(A, mode='reduced')
    
    true_cols = [int(in_str, local_dim) for in_str in in_strings]
    for col in true_cols: 
        Q[:,col] = A[:,col]
    return Q

def unitary_embedding(in_strings, out_states, local_dim=2): 
    # deprecated
    print("This function is deprecated. Use construct_unitary instead.")

    n = len(out_states[0])
    
    in_states = [string_to_sv(in_str, local_dim) for in_str in in_strings]
    #A = (np.array(out_states).conj().T @ np.array(in_states))
    A = (np.array(out_states).T @ np.array(in_states))
    
    true_rows = [int(in_str, local_dim) for in_str in in_strings]
    space = scipy.linalg.null_space(A.T)
    
    A_row = 0
    space_row = 0
    while space_row < n-len(in_strings): 
        if A_row not in true_rows: 
            A[:,A_row] = space[:,space_row]
            space_row += 1
        A_row += 1
        
    return A 

def k_body_z(n, k): 
    ops = []
    for indices in combinations(range(n), k): 
        name = "".join("Z" if i in indices else "I" for i in range(n))
        mat = embed_operator(n, list(indices), [pauli_Z] * k, dense=True)
        ops.append((name, mat))
    return ops

def rz_ansatz(n, params, body=1): 
    gate_list = [RZ(i, params[i]) for i in range(n)]
    for k in range(2,body+1):
        generators = k_body_z(n, k)
        gates = []
        for i, (name, mat) in enumerate(generators):
            u = expm(-1.j * params[len(gate_list)+i] * mat)
            gates.append(Gate('R'+name, np.arange(n), u))
        gate_list += gates

    return Circuit(gate_list, n)

def random_circ(n, depth): 
    skeleton = brickwall_skeleton(n,depth)
    num_cnots = len(skeleton)
    num_params = (num_cnots * 6) + (n * 3)
    return dressed_cnot_circ(skeleton, n, np.random.rand(num_params))



from .tn import *
from .initializations import *
from .statevector_sim import *
from .environments import *
from .density_matrix_sim import *
from .mps_sim import *
from .mpdo_sim import *
from .monte_carlo_sim import *


from qaravan.core.base_sim import BaseSim
from qaravan.core.utils import string_to_sv, pretty_print_dm
from qaravan.core.gates import SuperOp

import numpy as np
from ncon_torch import ncon

class DensityMatrixSim(BaseSim):
    def __init__(self, circ, init_state=None, nm=None):
        super().__init__(circ, init_state=init_state, nm=nm)    

    def initialize_state(self):
        """ 
        internal state is a rank-2n tensor with local dimension inherited from the circuit 
        init_state can be provided either as a tensor, a density matrix or a bitstring
        """
        if self.init_state is None: 
            dm = np.zeros((self.local_dim**self.num_sites, self.local_dim**self.num_sites))
            dm[0,0] = 1.0
            self.state = dm.reshape([self.local_dim]*self.num_sites*2)
        
        elif type(self.init_state) == np.ndarray:
            if len(self.init_state.shape) == 2:
                self.state = self.init_state.reshape([self.local_dim]*self.num_sites*2) 
            elif len(self.init_state.shape) == self.num_sites*2:
                self.state = self.init_state
            else:
                raise ValueError("init_state must be a rank-2 tensor or a rank-2n tensor.")

        elif type(self.init_state) == str: 
            sv = string_to_sv(self.init_state, self.circ.local_dim)
            dm = np.outer(sv, sv.conj())
            self.state = dm.reshape([self.local_dim]*self.num_sites*2)

        else:
            raise ValueError("init_state must be either a numpy array or a bitstring.") 

    def apply_gate(self, gate):
        """ not sure if this is restricted to 1- and 2-qubit gates """
        superop = gate.matrix if type(gate) == SuperOp else gate.to_superop()
        gate_indices, state_indices = generate_gate_indices(gate.indices, self.num_sites), generate_state_indices(gate.indices, self.num_sites)
        self.state = ncon((superop, self.state), (gate_indices, state_indices))

    def measure(self, meas_sites):
        """ should just use local_expectation """
        raise NotImplementedError("Measurement not yet implemented for density matrix simulator.")
    
    # TODO: allo local_ops to have span greater than 1
    def local_expectation(self, local_ops):
        """ local_ops is a list of local operators, one for each site """
        self.run(progress_bar=False)
        for i in range(self.num_sites):
            state_indices = [-(j+1) for j in range(2*self.num_sites)] 
            state_indices[i] = 1
            self.state = ncon((local_ops[i], self.state), ([-(i+1),1], state_indices))

        dm = self.state.reshape(self.local_dim**self.num_sites, self.local_dim**self.num_sites)
        return np.trace(dm).real
    
    def global_expectation(self, global_op):
        if not self.ran: 
            self.run(progress_bar=False)
        dm = self.state.reshape(self.local_dim**self.num_sites, self.local_dim**self.num_sites)
        return np.trace(dm @ global_op).real
    
    def __str__(self):
        dm = self.state.reshape(self.local_dim**self.num_sites, self.local_dim**self.num_sites)
        return pretty_print_dm(dm, self.local_dim)
    
    def get_density_matrix(self):
        return self.state.reshape(self.local_dim**self.num_sites, self.local_dim**self.num_sites)

def generate_gate_indices(locs, n): 
    numbers = [i for i in range(1,2*len(locs)+1)]
    contracted = sorted(numbers)[-len(numbers)//2:] + sorted(numbers)[:len(numbers)//2]
    
    numbers = [(int(loc)+1) for loc in locs] + [(int(loc)+n+1) for loc in locs]
    largest, smallest = sorted(numbers)[-len(numbers)//2:], sorted(numbers)[:len(numbers)//2]
    uncontracted = [-i for i in largest + smallest]
    return uncontracted + contracted

def generate_state_indices(locs, n):
    state_indices = [-i for i in range(1, 2 * n + 1)]
    pos = 1
    for i, loc in enumerate(locs):
        state_indices[loc] = pos
        state_indices[loc + n] = pos + len(locs)
        pos += 1
    return state_indices

def random_dm(n): 
    a = np.random.rand(2**n,2**n) + 1.j * np.random.rand(2**n,2**n)
    a = a@a.conj().T
    return a/np.trace(a)

def all_zero_dm(n, local_dim=2): 
    dm = np.zeros((local_dim**n, local_dim**n))
    dm[0,0] = 1.0
    return dm 

def partial_trace(dm, sites, local_dim=2): 
    system_size = int(np.log(dm.shape[0]) / np.log(local_dim))
    sites = sorted(sites)

    if len(sites) == 0: 
        return np.trace(dm).real
    
    if len(sites) == system_size:
        return dm 
    
    else: 
        dm = dm.reshape([local_dim] * system_size * 2)
        contraction = []
        counter = 0 
        for i in range(system_size):
            if i not in sites:
                counter += 1
                contraction.append(counter)

        contraction = contraction * 2 
        counter = -1
        for site in sites:
            contraction.insert(site, counter)
            counter -= 1

        for site in sites:
            contraction.insert(site + system_size, counter)
            counter -= 1

        return ncon([dm], contraction).reshape(local_dim**len(sites), local_dim**len(sites))

def expectation(dm, op): 
    return np.trace(dm @ op).real

def rdm_from_sv_full(sv, sites, local_dim=2): 
    """ Deprecated version of rdm_from_sv that uses the full state vector."""
    dm = np.outer(sv, sv.conj())
    return partial_trace(dm, sites, local_dim=local_dim)


from .tn import *
from numpy.random import uniform, normal
from sympy.combinatorics import Permutation

######################################
################ TI ##################
######################################

def one_copy_projector_ti_subspace(local_dimension, number_sites):
    """ returns r**number_sites where r is the effective local dimension """ 
    symmetry = np.concatenate([np.arange(number_sites, dtype=int), np.arange(number_sites, dtype=int)])
    permutation_list = [symmetry[i : i + number_sites] for i in range(number_sites)]
    length_permutation_list = len(permutation_list)
    cycle_counts = np.array([Permutation(permutation_list[i]).cycles for i in range(length_permutation_list)], dtype=int)
    return_value = np.sum(float(local_dimension) ** cycle_counts) / len(permutation_list)
    return return_value

class TI_MPS(MPS):
    def __init__(self, site, num_sites):
        super().__init__([site] * num_sites)
    
    def overlap(self, other, scaled=False): 
        """ if scaled, multiplies every transfer matrix by sqrt(r) where d is the effective local dimension """
        if scaled:
            scale = np.sqrt(one_copy_projector_ti_subspace(self.local_dim, self.num_sites)**(1/self.num_sites)) 
        else: 
            scale = 1.0
            
        tensor = ncon((self.sites[0], other.sites[0].conj()), ([-1,-3,1],[-2,-4,1])) * scale
        for i in range(1,self.num_sites):
            tensor = ncon((tensor, self.sites[i]), ([-1,-2,1,-3],[1,-4,-5]))
            tensor = ncon((tensor, other.sites[i].conj()), ([-1,-2,1,-3,2],[1,-4,2])) * scale
        return np.trace(tensor.reshape(tensor.shape[0] * tensor.shape[1], tensor.shape[2] * tensor.shape[3]))

######################################
############### HAAR #################
######################################
    
def random_unitary(n, local_dim=2): 
    # generates a random local_dim^n x local_dim^n unitary matrix
    a = np.random.rand(local_dim**n,local_dim**n) + 1j*np.random.rand(local_dim**n,local_dim**n)
    q,r = np.linalg.qr(a)
    d = np.diagonal(r)
    ph = d/np.abs(d)
    q *= ph
    return q

def haar_random_isometry(l_chi, r_chi=None, local_dim=2):
    r_chi = l_chi if r_chi is None else r_chi
    size = local_dim * l_chi
    temp = np.random.rand(size, size) + 1j * np.random.rand(size, size)
    temp = temp + temp.conj().T 
    return np.linalg.eigh(temp)[1][:, :r_chi]

#######################################
############## RMPS ###################
#######################################

def open_rmps_staggered(num_sites, chi, local_dim=2, distrib='isometric', param=1.0):
    sites = []
    for i in range(num_sites): 
        l_chi = min(local_dim**i, chi, local_dim**(num_sites-i))
        r_chi = min(local_dim**(i+1), chi, local_dim**(num_sites-1-i))
        shape = (l_chi, r_chi, local_dim)
        
        if distrib == 'isometric':
            site = haar_random_isometry(*shape).reshape(l_chi, local_dim, r_chi).transpose(0,2,1)
        elif distrib == 'uniform': 
            site = uniform(-1, 1, shape) + 1.j*uniform(-1, 1, shape)
        elif distrib == 'gaussian':
            site = normal(0, param, shape) + 1.j*normal(0, param, shape)
        else: 
            raise ValueError(f"{distrib} is not a valid distribution for site tensors")
        sites.append(site)
        
    mps = MPS(sites)
    if distrib != 'isometric': 
        mps.normalize()
    return mps 

def open_rmps_even(num_sites, chi, local_dim=2, distrib='isometric', param=1.0):
    sites = []
    for _ in range(num_sites):
        if distrib == 'isometric':
            site = haar_random_isometry(chi, chi, local_dim).reshape(chi, local_dim, chi).transpose(0,2,1) 
        elif distrib == 'uniform':
            site = uniform(-1, 1, (chi, chi, local_dim)) + 1.j * uniform(-1, 1, (chi, chi, local_dim))
        elif distrib == 'gaussian':
            site = normal(0, param, (chi, chi, local_dim)) + 1.j * normal(0, param, (chi, chi, local_dim))
        else: 
            raise ValueError(f"{distrib} is not a valid distribution for site tensors")
        sites.append(site)

    edge = np.zeros(chi)
    edge[0] = 1.0
    sites[0] = ncon((edge, sites[0]), ([1], [1,-1,-2]))[np.newaxis, :, :]
    sites[-1] = ncon((sites[-1], edge), ([-1,1,-2], [1]))[:, np.newaxis, :]
    mps = MPS(sites)
    mps.normalize()
    return mps
    
def periodic_rmps(num_sites, chi, local_dim=2, distrib="isometric", param=1.0): 
    sites = []
    for _ in range(num_sites):
        if distrib == 'isometric':
            site = haar_random_isometry(chi, chi, local_dim).reshape(chi, local_dim, chi).transpose(0,2,1)
        elif distrib == 'uniform':
            site = uniform(-1, 1, (chi, chi, local_dim)) + 1.j * uniform(-1, 1, (chi, chi, local_dim))
        elif distrib == 'gaussian': 
            site = normal(0, param, (chi, chi, local_dim)) + 1.j * normal(0, param, (chi, chi, local_dim))
        else: 
            raise ValueError(f"{distrib} is not a valid distribution for site tensors")
        sites.append(site)

    mps = MPS(sites)
    mps.normalize()
    return mps

def ti_rmps(num_sites, chi, local_dim=2, distrib="isometric", param=1.0):
    site = haar_random_isometry(chi, chi, local_dim).reshape(chi, local_dim, chi).transpose(0,2,1)
    mps = TI_MPS(site, num_sites)
    mps.normalize()
    return mps

###################################################
############### MPOs ##############################
###################################################

def rand_mpo(n, chi, boundary='open', local_dim=2):
    if boundary == 'open': 
        sites = [np.random.rand(1,chi,local_dim,local_dim)] + [np.random.rand(chi,chi,local_dim,local_dim)
                     for _ in range(n-2)] + [np.random.rand(chi,1,local_dim,local_dim)]
    elif boundary == 'periodic': 
        sites = [np.random.rand(chi,chi,local_dim,local_dim) for _ in range(n)]
    else: 
        raise ValueError(f"Invalid boundary: {boundary}")
        
    return MPO(sites)

def all_zero_mps(n, local_dim=2):
    vec = np.zeros(local_dim)
    vec[0] = 1.0
    return MPS([vec[np.newaxis,np.newaxis,:]] * n)

def string_to_mps(string, local_dim): 
    sites = []
    for s in string: 
        vec = np.zeros(local_dim)
        vec[int(s)] = 1.0
        sites.append(vec[np.newaxis, np.newaxis, :])
    return MPS(sites)

def all_zero_mpdo(n, local_dim=2):
    mat = np.zeros((local_dim, local_dim))
    mat[0,0] = 1.0
    return MPDO([mat[np.newaxis,np.newaxis,:,:] for i in range(n)])


from .statevector_sim import *

class MonteCarloSim(BaseSim):
    """ 
    noisy simulation for channels that can be thought of as 
    a pseudo-probability distribution over unitaries applied to statevector
    nm must have a method 'sample' that returns a list of Gates
    main output of simulator is expectation values
    unlike other simulators it doesn't store internal states  
    """   
    def __init__(self, circ, init_state=None, nm=None):
        super().__init__(circ, init_state=init_state, nm=nm)  

    def circuit_sample(self): 
        circ = copy.deepcopy(self.circ)
        if not circ.built: 
            circ.decompose()
            circ.construct_layers()
        
        if self.nm is not None: 
            noise_layers = [self.nm.sample() for _ in range(len(circ.layers))]
            circ.layers = [
                circ.layers[i//2] if i % 2 == 0 else noise_layers[i//2]
                for i in range(len(circ.layers) * 2)
            ]

        circ.built = True
        return circ
    
    def expectation_sample(self, obs): 
        """ for now obs is a matrix of correct dimension """
        circ = self.circuit_sample()
        sim = StatevectorSim(circ, init_state=self.init_state)
        sim.run(progress_bar=False)
        sv = sim.get_statevector()
        return np.vdot(sv, obs @ sv).real


from .noise import *
from .gates import *
from .param_gates import *
from .paulis import *
from .utils import *
from .base_sim import *
from .lattices import *
from .hamiltonians import *
from .circuits import *


from qaravan.core.gates import SuperOp, ID
from qaravan.core.noise import gate_time 
from .initializations import *
import copy
from tqdm.notebook import tqdm

###### nsim.py ########

##### A simpler way to do all this might be to just reshape the MPO into MPS and then apply with @ instead of ncon

def layer_to_superop(layer, span, start_idx, nm=None, name="S", time=0.0): 
    local_dim = layer[0].get_local_dim()
    
    # first we sort and pad the layer
    current_indices = set(idx for gate in layer for idx in gate.indices)
    missing_indices = set(range(span)) - current_indices
    id_gates = [ID(local_dim=local_dim, time=0.0, indices=[idx]) for idx in missing_indices]
    layer = sorted(layer + id_gates, key=lambda x: x.indices[0])
    
    # next we tensor product gates and reorder indices
    arrays = [g.to_superop(nm=nm) for g in layer]
    
    result = arrays[0]
    for i in range(1, len(arrays)):
        result = np.tensordot(result, arrays[i], axes=0)
        
    shapes = [ar.shape for ar in arrays]
    cumsum = [0] + [sum(len(s) for s in shapes[:i]) for i in range(1, len(shapes))]
    indices = [list(range(c, c + len(s))) for c, s in zip(cumsum, shapes)]
    quarters = [sum([sublist[i*len(sublist)//4:(i+1)*len(sublist)//4] for sublist in indices], []) 
                                                                                for i in range(4)]
    new_order = [item for sublist in quarters for item in sublist]
    return SuperOp(name, span, result.transpose(*new_order), start_idx=start_idx, time=time)

def noise_superop(layer, span, start_idx, nm, name="N"): 
    layer_time = max([gate_time(gate, nm) for gate in layer])
    gate_times = [0.0 for i in range(span)]
    for i in range(span): 
        for gate in layer: 
            if i in gate.indices:
                gate_times[i] = gate_time(gate, nm)

    noise_layer = [ID(nm.local_dim, max(nm.coupling*layer_time, gate_times[i]), i) 
                   for i in range(span)]
    return layer_to_superop(noise_layer, span, start_idx, nm=nm, name=name, time=np.array(gate_times))

def prod(superop1, superop2, name="S"): 
    """ returns the composition of two superoperators """
    num_indices = len(superop1.shape)
    matrix = np.tensordot(superop1.matrix, superop2.matrix, axes=(range(num_indices//2, num_indices), 
                                                                range(num_indices//2)))
    return SuperOp(name, superop1.span, matrix, start_idx=superop1.start_idx, time=superop1.time+superop2.time)

def circ_to_superop(circ, span, start_idx, nm=None): 
    new_circ = circ.build()

    superop_list = []
    for layer in new_circ.layers: 
        superop_list.append(layer_to_superop(layer, span, start_idx))
        if nm is not None: 
            noise_layer = noise_superop(layer, span, start_idx, nm)
            superop_list.append(noise_layer)

    return reduce(prod, superop_list[::-1])

def one_local(superop, site): 
    """ 
    applies single qudit superoperators; 
    superop.shape = (two left indices - top to bottom, two right indices - top to bottom)
    site.shape = (left_bond, right_bond, top_spin, bottom_spin=d)
    output.shape = (left_bond, right_bond, top_spin, bottom_spin)
    """
    output = ncon((superop, site), ([-4,-3,2,1],[-1,-2,1,2]))
    return output

def two_local(superop, site):
    """ 
    applies two qudit superoperators; 
    superop.shape = (four left indices - top to bottom, four right indices - top to bottom)
    site.shape = (left_bond, first_top_spin, first_bottom_spin, 
                    right_bond, second_top_spin, second_bottom_spin)
    output.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    left_bond, right_bond)
    """
    output = ncon((superop, site), ([-2,-4,-1,-3,2,4,1,3],[-5,1,2,-6,3,4])) 
    return output

def three_local(superop, site):
    """ 
    applies three qudit superoperators; 
    superop.shape = (six left indices - top to bottom, six right indices - top to bottom)
    site.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin
                    third_top_spin, third_bottom_spin, left_bond, right_bond)
    output.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    third_top_spin, third_bottom_spin, left_bond, right_bond)
    """
    output = ncon((superop, site), ([-2,-4,-6,-1,-3,-5,2,4,6,1,3,5],[1,2,3,4,5,6,-7,-8])) 
    return output

def four_local(superop, site):
    """ 
    applies four qudit superoperators; 
    superop.shape = (eight left indices - top to bottom, eight right indices - top to bottom)
    site.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin
                    third_top_spin, third_bottom_spin, fourth_top_spin, fourth_bottom_spin, left_bond, right_bond)
    output.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    third_top_spin, third_bottom_spin, fourth_top_spin, fourth_bottom_spin, left_bond, right_bond)
    """
    output = ncon((superop, site), ([-2,-4,-6,-8,-1,-3,-5,-7,2,4,6,8,1,3,5,7],[1,2,3,4,5,6,7,8,-9,-10])) 
    return output

def superop_action(mpdo, gate, nm=None): 
    """ applies a 1- or 2-qubit superoperator to density matrix represented as MPO """
    sites = mpdo.sites
    inds = gate.indices
    superop = gate.matrix if type(gate) == SuperOp else gate.to_superop(nm)
    
    if len(inds) == 1:
        sites[inds[0]] = one_local(superop, sites[inds[0]])
        
    elif len(inds) == 2:
        data = ncon((sites[inds[0]], sites[inds[1]]), ([-1,1,-2,-3],[1,-4,-5,-6]))
        data = two_local(superop, data)
        sites[inds[0]], sites[inds[1]] = decompose_site(data)            
    
    elif len(inds) == 3: 
        data = ncon((sites[inds[0]], sites[inds[1]], sites[inds[2]]), ([-7,1,-1,-2],[1,2,-3,-4],[2,-8,-5,-6]))
        data = three_local(superop, data)
        sites[inds[0]], sites[inds[1]], sites[inds[2]] = decompose_three_site(data)
    
    elif len(inds) == 4: 
        data = ncon((sites[inds[0]], sites[inds[1]], sites[inds[2]], sites[inds[3]]), 
                                            ([-9,1,-1,-2],[1,2,-3,-4],[2,3,-5,-6],[3,-10,-7,-8]))
        data = four_local(superop, data) 
        sites[inds[0]], sites[inds[1]], sites[inds[2]], sites[inds[3]] = decompose_four_site(data)
    
    else: 
        raise NotImplementedError(f"{len(inds)}-site gates have not been implemented") 
    
    return mpdo

def nt_sim(circ, **kwargs): 
    """ 
    noisy MPO simulation of circuit represented as gate_list 
    initial state is assumed to be the all zero state unless otherwise stated
    truncation is controlled by msvr and max_dim
    """
    init_mpdo, nm, msvr, max_dim, dm, debug, quiet = (kwargs.get('init_mpdo'), kwargs.get('nm'), kwargs.get('msvr'), 
            kwargs.get('max_dim'), kwargs.get('dm',False), kwargs.get('debug',False), kwargs.get('quiet',True))
    
    mpdo = all_zero_mpdo(circ.num_sites, circ.local_dim) if init_mpdo is None else copy.deepcopy(init_mpdo)
    circ = circ.build(nm)
    
    it = circ.layers if quiet else tqdm(circ.layers)
    for i, layer in enumerate(it): 
        for gate in layer: 
            mpdo = superop_action(mpdo, gate, nm)
            if debug: 
                pretty_print_dm(mpdo.to_matrix(), circ.local_dim)
                print(mpdo.get_skeleton())
                print("")
        
        mpdo = mpdo.compress(msvr, max_dim)
    
    if circ.meas_qubits is not None: 
        rdm = mpdo.partial_trace(circ.meas_qubits)
            
    return mpdo.to_matrix() if dm else mpdo


from qaravan.core import embed_operator, pauli_Z
from qaravan.tensorQ import all_zero_dm, all_zero_sv, StatevectorSim, DensityMatrixSim, MonteCarloSim
import numpy as np
from tqdm import tqdm

def magnetization(n):
    """ return the properly normalized total magnetization operator on n qubits"""
    mz = sum(embed_operator(n, [i], [pauli_Z], dense=True) for i in range(n))
    return mz / np.sqrt(n * 2**n)

def trotter_sv_sim(ham, step_size, max_steps, op):
    """ run a statevector simulation of the Trotter circuit for a given Hamiltonian """
    circ = ham.trotter_circ(step_size, 1)
    cur_sv = all_zero_sv(ham.num_sites, dense=True)
    
    exp_list = []
    for _ in tqdm(range(max_steps)):
        sim = StatevectorSim(circ=circ, init_state=cur_sv)
        sim.run(progress_bar=False)
        cur_sv = sim.get_statevector()
        exp = np.vdot(cur_sv, op @ cur_sv).real
        exp_list.append(exp)
    return exp_list

def trotter_dm_sim(ham, step_size, max_steps, channel, op): 
    """ run a density matrix simulation of the Trotter circuit for a given Hamiltonian """
    circ = ham.trotter_circ(step_size, 1)
    cur_dm = all_zero_dm(ham.num_sites)
    
    exp_list = []
    for _ in tqdm(range(max_steps)):
        sim = DensityMatrixSim(circ=circ, nm=channel, init_state=cur_dm)
        sim.run(progress_bar=False)
        cur_dm = sim.get_density_matrix()
        noisy_exp = np.trace(op @ cur_dm).real
        exp_list.append(noisy_exp)
    return exp_list

def trotter_dm_sim_uncached(ham, step_size, max_steps, channel, op): 
    """ slow version without caching the state """
    mag_list = []
    for num_steps in tqdm(range(1, max_steps + 1)):
        circ = ham.trotter_circ(step_size, num_steps)
        sim = DensityMatrixSim(circ=circ, nm=channel)
        sim.run(progress_bar=False)
        dm = sim.get_density_matrix()
        noisy_exp = np.trace(op @ dm).real
        mag_list.append(noisy_exp)
    return mag_list

def trotter_mc_sim_uncached(ham, step_size, max_steps, channel, op, num_samples=1000): 
    """ slow version without caching the state """
    mag_list_mc = []
    for num_steps in tqdm(range(1, max_steps + 1)):
        circ = ham.trotter_circ(step_size, num_steps)
        sim = MonteCarloSim(circ=circ, nm=channel)
        mag_samples = [sim.expectation_sample(op) for _ in tqdm(range(num_samples))]
        mag_list_mc.append(np.mean(mag_samples))
    return mag_list_mc

def trotter_mc_sim(ham, step_size, max_steps, channel, op, num_samples=1000):
    """ run a Monte Carlo simulation of the Trotter circuit for a given Hamiltonian """
    circ = ham.trotter_circ(step_size, 1)
    sim = MonteCarloSim(circ=circ, nm=channel)    
    cur_sv_list = [all_zero_sv(ham.num_sites, dense=True) for _ in range(num_samples)]
    
    mag_list_mc = []
    for _ in tqdm(range(max_steps)):
        layer_samples = [sim.circuit_sample() for _ in range(num_samples)]
        mag_samples = []
        for i in range(num_samples):
            sv_sim = StatevectorSim(layer_samples[i], init_state=cur_sv_list[i])
            sv_sim.run(progress_bar=False)
            new_sv = sv_sim.get_statevector()
            cur_sv_list[i] = new_sv
            mag_samples.append(np.vdot(new_sv, op @ new_sv).real)

        mag_list_mc.append(np.mean(mag_samples))
    return mag_list_mc


from qaravan.core import Circuit, Gate
from typing import Tuple, List
import stim
import itertools

def cnot_conversion(gate: Gate) -> Tuple[str, List[int]]:
    """ convert CNOT gate to Stim format """
    name = "CNOT"
    if gate.bottom_heavy: 
        indices = gate.indices
    else:
        indices = gate.indices[::-1]
    return name, indices

def qaravan_to_stim(circ: Circuit) -> stim.Circuit:
    """ convert a Qaravan circuit to a Stim circuit """
    stim_circ = stim.Circuit()
    for gate in circ:
        try: 
            if gate.name == "CNOT2": 
                stim_circ.append(*cnot_conversion(gate)) 
            else: 
                stim_circ.append(gate.name, gate.indices)
        except Exception:
            raise NotImplementedError(f"Gate {gate} not implemented in Stim conversion.")
    
    return stim_circ

def group_from_generators(generators: List[stim.PauliString]) -> List[stim.PauliString]:
    """ create a group from a list of generators """
    stabilizer_group = []
    for bits in itertools.product([0, 1], repeat=len(generators)):
        p = stim.PauliString(len(generators[0])) 
        for i, bit in enumerate(bits):
            if bit:
                p *= generators[i]
        stabilizer_group.append(p)

    return stabilizer_group

def is_stabilized_by(circuit: stim.Circuit, pauli: stim.PauliString) -> bool:
    """ check if a circuit stabilizes a Pauli string """
    sim = stim.TableauSimulator()
    sim.do_circuit(circuit)
    return sim.peek_observable_expectation(pauli) == 1

def string_to_stim(pstr: str) -> stim.PauliString:
    """ convert a string representation of a Pauli string to a stim.PauliString """
    return stim.PauliString(pstr.replace('x', 'X').replace('y', 'Y').replace('z', 'Z').replace('i', 'I'))

def stim_to_string(pauli: stim.PauliString) -> str:
    """ convert a stim.PauliString to a string representation """
    return str(pauli)[1:].replace('X', 'x').replace('Y', 'y').replace('Z', 'z').replace('_', 'i')


import numpy as np
from qaravan.tensorQ import *
from qaravan.core import *
from qaravan.applications import *
import torch
from collections import Counter

def test_statevector_sim():
    gate_list = [
        H(0),               
        CNOT([1, 0], 3),    
        CNOT([2, 1], 3)    
    ]
    circ = Circuit(gate_list, 3)
    
    sim = StatevectorSim(circ)
    sim.run()
    statevector = sim.get_statevector()

    expected = np.zeros(8, dtype=complex)
    expected[0] = 1 / np.sqrt(2)
    expected[-1] = 1 / np.sqrt(2)

    assert np.allclose(statevector, expected, atol=1e-7), "Statevector is not the expected GHZ state"

def test_construct_unitary():
    in_strings = ['00', '01', '10']
    local_dim = 2

    unitary = random_unitary(local_dim**2)
    out_states = [unitary @ string_to_sv(in_str, local_dim) for in_str in in_strings]
    Q = construct_unitary(in_strings, out_states, local_dim)

    assert np.allclose(Q.conj().T @ Q, np.eye(4), atol=1e-7), "Q is not unitary (Qd Q != I)."
    assert np.allclose(Q @ Q.conj().T, np.eye(4), atol=1e-7), "Q is not unitary (Q Qd != I)."

    for st in in_strings:
        original = unitary @ string_to_sv(st, local_dim)
        constructed = Q @ string_to_sv(st, local_dim)
        assert np.allclose(constructed, original, atol=1e-7), (
            f"Constructed Q does not match reference unitary for input state '{st}'."
        )

def test_contract_and_decimate():
    # Test contract_sites produces the correct shape for MPS sites
    sites = [np.random.rand(4,3,2)] + [np.random.rand(3,3,2)] * 1 + [np.random.rand(3,5,2)]
    c_site = contract_sites(sites)
    assert c_site.shape == (4, 5, 2**3), "contract_sites failed for MPS sites"

    # Test contract_sites produces the correct shape for MPDO sites
    sites = [np.random.rand(4,3,2,2)] + [np.random.rand(3,3,2,2)] * 1 + [np.random.rand(3,5,2,2)]
    c_site = contract_sites(sites)
    assert c_site.shape == (4, 5, 2**6), "contract_sites failed for MPDO sites"

    # Test decimate produces the correct shape and correct tensors for MPS sites
    sites = [np.random.rand(4,3,2)] + [np.random.rand(3,3,2)] * 1 + [np.random.rand(3,5,2)]
    c_site = contract_sites(sites)
    dec_sites = decimate(c_site, 2)
    c_site2 = contract_sites(dec_sites)
    assert np.allclose(c_site, c_site2), "decimate failed for MPS sites"
    assert all(s.shape == d.shape for s, d in zip(sites, dec_sites)), "Shapes do not match for MPS sites"

    # Test decimate produces the correct shape and correct tensors for MPDO sites
    sites = [np.random.rand(4,3,2,2)] + [np.random.rand(3,3,2,2)] * 1 + [np.random.rand(3,5,2,2)]
    c_site = contract_sites(sites)
    dec_sites = decimate(c_site, 2*2)  # super local_dim is square of local_dim
    dec_sites = [s.reshape(s.shape[0], s.shape[1], 2, 2) for s in dec_sites]
    c_site2 = contract_sites(dec_sites)
    assert np.allclose(c_site, c_site2), "decimate failed for MPDO sites"
    assert all(s.shape == d.shape for s, d in zip(sites, dec_sites)), "Shapes do not match for MPDO sites"

def test_circuit_copy_with_autograd_gate():
    theta = torch.randn(1, dtype=torch.float64, requires_grad=True)
    mat = torch.matrix_exp(-1j * theta * torch.tensor([[0., 1.], [1., 0.]], dtype=torch.complex128))
    assert mat.requires_grad

    gate = Gate('Rx', [0], mat)
    circ = Circuit([gate], n=1)
    circ_copy = circ.copy()
    
    sv = torch.tensor([1.0, 0.0], dtype=torch.complex128)
    sv_out = torch.matmul(gate.matrix, sv)
    loss = torch.real(torch.dot(sv_out.conj(), sv_out))
    loss.backward()
    assert theta.grad is not None
    print("Test passed: Circuit with autograd-compatible gate copied successfully.")

def test_ti_rmps_generation():
    num_sites = 4
    chi = 2

    rmps = ti_rmps(num_sites, chi)
    ref = rmps.sites[0]
    for site in rmps.sites:
        assert np.allclose(site, ref), "Not all tensors in rmps.sites are the same"
    
    print("Test passed: TI-RMPS generation successful with translation invariant tensors.")

def test_measurement_on_statevector():
    sv = (
        np.sqrt(0.3) * string_to_sv('0001') +
        np.sqrt(0.45) * string_to_sv('1110') +
        np.sqrt(0.25) * string_to_sv('1011')
    )
    counts = Counter()
    num_samples = 5000
    for _ in range(num_samples):
        outcome = measure_sv(sv, [0, 1])
        counts[outcome] += 1

    assert abs(counts['00'] / num_samples - 0.3) < 0.05, f"Expected ~0.3, got {counts['00'] / num_samples}"
    assert abs(counts['11'] / num_samples - 0.45) < 0.05, f"Expected ~0.45, got {counts['11'] / num_samples}"
    assert abs(counts['10'] / num_samples - 0.25) < 0.05, f"Expected ~0.25, got {counts['10'] / num_samples}"

def test_measure_and_collapse_sv():
    sv = (
        np.sqrt(0.3) * string_to_sv('0001') +
        np.sqrt(0.45) * string_to_sv('1110') +
        np.sqrt(0.25) * string_to_sv('1011')
    )

    num_trials = 50
    for _ in range(num_trials):
        meas_indices = [0, 2]
        outcome, new_sv = measure_and_collapse_sv(sv, meas_indices)

        if outcome == '00':
            assert np.allclose(new_sv, string_to_sv('01')), f"Collapsed state mismatch for outcome '00'"
        elif outcome == '11':
            true_outcome = (
                np.sqrt(0.45 / (0.45 + 0.25)) * string_to_sv('10') +
                np.sqrt(0.25 / (0.45 + 0.25)) * string_to_sv('01')
            )
            assert np.allclose(new_sv, true_outcome, atol=1e-7), f"Collapsed state mismatch for outcome '11'"
        else:
            assert False, f"Unexpected measurement outcome: {outcome}"


import numpy as np
import matplotlib.pyplot as plt
from collections import OrderedDict

class Lattice:
    def __init__(self, Lx, Ly, periodic_x=False, periodic_y=False):
        self.Lx = Lx
        self.Ly = Ly
        self.periodic_x = periodic_x
        self.periodic_y = periodic_y
        self.atom_positions = {}
        self.cells = set()
        self.num_sites = None

class LinearLattice(Lattice): 
    def __init__(self, L, periodic=False):
        super().__init__(L, 1, periodic_x=periodic, periodic_y=False)
        self.atom_positions = {i: (i, 0) for i in range(L)}
        self.cells = set([i for i in range(L)])
        self.num_sites = L

    def get_connections(self, locality='nn'): 
        if locality == 'nn':
            nn_pairs = [(i, i + 1) for i in range(self.Lx - 1)]
            if self.periodic_x:
                nn_pairs.append((self.Lx - 1, 0))
            return nn_pairs
        
        elif locality == 'nnn': 
            nnn_pairs = [(i, i + 2) for i in range(self.Lx - 2)]
            if self.periodic_x:
                nnn_pairs.append((self.Lx - 2, 0))
                nnn_pairs.append((self.Lx - 1, 1))
            return nnn_pairs
        
        elif locality == 'a2a': 
            a2a_pairs = [(i,j) for i in range(self.Lx) for j in range(i+1, self.Lx)]
            return a2a_pairs

        else:
            raise ValueError("Unsupported locality type")

class ToricLattice:
    def __init__(self, Lx, Ly):
        """ 
        qubits on horizontal edges are indexed first (left to right, top to bottom),
        followed by vertical edges (left to right, top to bottom).
        Edges connecting boundary vertices are also decorated with qubits due to periodicity.
        (x, y, 'h'/'v') refers to an edge starting from vertex (x, y) and going right or down.
        """
        self.Lx = Lx
        self.Ly = Ly
        self.num_sites = 2 * Lx * Ly

        self.edge_map = {}          # (x, y, 'h'/'v') → index
        self.index_map = {}         # index → (x, y, 'h'/'v')
        self.vertex_terms = []      # list of 4-site lists
        self.plaquette_terms = []   # list of 4-site lists

        self._build_edges()
        self._build_vertex_terms()
        self._build_plaquette_terms()

    def _build_edges(self):
        idx = 0
        for y in range(self.Ly):
            for x in range(self.Lx):
                self.edge_map[(x, y, 'h')] = idx
                self.index_map[idx] = (x, y, 'h')
                idx += 1
        for y in range(self.Ly):
            for x in range(self.Lx):
                self.edge_map[(x, y, 'v')] = idx
                self.index_map[idx] = (x, y, 'v')
                idx += 1

    def _build_vertex_terms(self):
        for y in range(self.Ly):
            for x in range(self.Lx):
                edges = [
                    self.edge_map[(x, y, 'h')],
                    self.edge_map[((x - 1) % self.Lx, y, 'h')],
                    self.edge_map[(x, y, 'v')],
                    self.edge_map[(x, (y - 1) % self.Ly, 'v')],
                ]
                self.vertex_terms.append(edges)

    def _build_plaquette_terms(self):
        for y in range(self.Ly):
            for x in range(self.Lx):
                edges = [
                    self.edge_map[(x, y, 'h')],
                    self.edge_map[((x + 1) % self.Lx, y, 'v')],
                    self.edge_map[(x, (y + 1) % self.Ly, 'h')],
                    self.edge_map[(x, y, 'v')],
                ]
                self.plaquette_terms.append(edges)

    def __getitem__(self, index):
        return self.index_map[index]

    def plot(self):
        raise NotImplementedError("Drawing function not implemented for ToricLattice")
    
class KagomeLattice:
    def __init__(self, row_layout, periodic_x=False, periodic_y=False):
        """
        General Kagome lattice constructed from a list of rows.
        Each row is a dict: {'num_cells': int, 'shift': float}
        Qubits live on A, B, C sites in triangles.
        """
        self.row_layout = row_layout
        self.periodic_x = periodic_x
        self.periodic_y = periodic_y

        self.site_map = OrderedDict()   # (i, j, 'A'/'B'/'C') → index
        self.index_map = {}             # index → (x, y, label)
        self.nn_pairs = set()
        self.triangle_terms = set()
        self.num_sites = 0

        self._place_sites()
        self._build_nn()
        self._build_triangles()

    def _place_sites(self):
        """Places qubit sites at triangle corners, shifted by row_layout."""
        index = 0
        triangle_basis = {
            'A': np.array([0.0, 0.0]),
            'B': np.array([0.5, 0.0]),
            'C': np.array([0.25, 0.5])
        }

        for j, row in enumerate(self.row_layout):
            shift_j = row.get('shift', 0.0)
            for i in range(row['num_cells']):
                for label, offset in triangle_basis.items():
                    x = i + offset[0] + shift_j
                    y = j + offset[1]
                    key = (i, j, label)
                    self.site_map[key] = index
                    self.index_map[index] = (x, y, label)
                    index += 1

        self.num_sites = index

    def _build_nn(self):
        """Build intra- and inter-triangle nearest-neighbor pairs with correct shift logic."""
        triangle_edges = [('A', 'B'), ('B', 'C'), ('C', 'A')]

        # Intra-triangle pairs
        for (i, j, label), idx in self.site_map.items():
            for a, b in triangle_edges:
                if label == a:
                    key_b = (i, j, b)
                    if key_b in self.site_map:
                        idx_b = self.site_map[key_b]
                        self.nn_pairs.add(tuple(sorted((idx, idx_b))))

        # Inter-triangle pairs
        for (i, j, label), idx in self.site_map.items():
            row = self.row_layout[j]
            Lx = row['num_cells']
            shift_j = row.get('shift', 0.0)

            if label == 'B':
                # B(i,j) → A(i+1,j)
                ni = (i + 1) % Lx if self.periodic_x else (i + 1)
                if 0 <= ni < Lx:
                    key = (ni, j, 'A')
                    if key in self.site_map:
                        self.nn_pairs.add(tuple(sorted((idx, self.site_map[key]))))

            elif label == 'C':
                nj = (j + 1) % len(self.row_layout) if self.periodic_y else (j + 1)
                if 0 <= nj < len(self.row_layout):
                    row_next = self.row_layout[nj]
                    shift_next = row_next.get('shift', 0.0)
                    Lx_next = row_next['num_cells']

                    delta = shift_next - shift_j
                    if delta > 0.0: # Row above is shifted right
                        key_A = (i, nj, 'A')
                        key_B = (i - int(delta/0.5), nj, 'B')
                    elif delta < 0.0: # Row above is shifted left
                        key_A = (i + int(-delta/0.5), nj, 'A')
                        key_B = (i, nj, 'B')
                    else:
                        continue

                    for key in [key_A, key_B]:
                        ki, kj, sub = key
                        if self.periodic_x:
                            ki = ki % Lx_next
                        if not (0 <= ki < Lx_next):
                            continue
                        full_key = (ki, kj, sub)
                        if full_key in self.site_map:
                            idx_b = self.site_map[full_key]
                            self.nn_pairs.add(tuple(sorted((idx, idx_b))))

    def _build_triangles(self):
        """Add in-cell and inter-row triangles with consistent orientation."""
        # Intra-cell ABC triangles
        for (i, j, _) in self.site_map:
            try:
                a = self.site_map[(i, j, 'A')]
                b = self.site_map[(i, j, 'B')]
                c = self.site_map[(i, j, 'C')]
                self.triangle_terms.add((a, b, c))  # counter-clockwise

            except KeyError:
                continue

        # Inter-row triangles formed between C(i,j) and row j+1
        for (i, j, label) in self.site_map:
            if label != 'C':
                continue

            shift_j = self.row_layout[j].get('shift', 0.0)
            nj = (j + 1) % len(self.row_layout) if self.periodic_y else j + 1
            if not (0 <= nj < len(self.row_layout)):
                continue

            shift_next = self.row_layout[nj].get('shift', 0.0)
            delta = shift_next - shift_j
            Lx_next = self.row_layout[nj]['num_cells']

            # Define triangle based on relative shift
            if delta > 0.0: 
                # Triangle: C(i,j), A(i, j+1), B(i-1, j+1)
                key_A = (i, nj, 'A')
                key_B = (i - int(delta/0.5), nj, 'B')
            elif delta < 0.0:
                # Triangle: C(i,j), A(i+1, j+1), B(i, j+1)
                key_A = (i + int(-delta/0.5), nj, 'A')
                key_B = (i, nj, 'B')
            else:
                continue

            # Index wrap + bounds
            triangle = []
            for key in [key_A, key_B]:
                ki, kj, sub = key
                if self.periodic_x:
                    ki = ki % Lx_next
                if not (0 <= ki < Lx_next):
                    break
                full_key = (ki, kj, sub)
                if full_key in self.site_map:
                    triangle.append(self.site_map[full_key])
            # Add C(i,j)
            triangle.insert(0, self.site_map[(i, j, 'C')])

            if len(triangle) == 3:
                self.triangle_terms.add(tuple(triangle))  # keep orientation consistent

    def __getitem__(self, index):
        return self.index_map[index]

    def get_connections(self, locality='nn'):
        if locality == 'nn':
            return list(self.nn_pairs)
        elif locality == 'triangle':
            return self.triangle_terms
        else:
            raise ValueError(f"Unsupported locality: {locality}")

    def plot(self, show_labels=True, show_triangles=True, figsize=(8, 6), save_path=None):
        _, ax = plt.subplots(figsize=figsize)
        colors = {'A': 'blue', 'B': 'green', 'C': 'red'}

        for idx, (x, y, label) in self.index_map.items():
            ax.scatter(x, y, s=100, color=colors[label], edgecolor='black', zorder=2)
            if show_labels:
                ax.text(x, y, str(idx), ha='center', va='center', color='white', fontsize=8)

        for i, j in self.nn_pairs:
            x1, y1, _ = self.index_map[i]
            x2, y2, _ = self.index_map[j]
            ax.plot([x1, x2], [y1, y2], 'k-', lw=1, zorder=1)

        if show_triangles:
            for a, b, c in self.triangle_terms:
                xa, ya, _ = self.index_map[a]
                xb, yb, _ = self.index_map[b]
                xc, yc, _ = self.index_map[c]
                ax.plot([xa, xb, xc, xa], [ya, yb, yc, ya], 'gray', lw=0.5, linestyle='--', zorder=0)

        ax.set_aspect('equal')
        ax.axis('off')
        plt.title("Kagome Lattice")
        if save_path:
            plt.savefig(save_path, bbox_inches='tight')

    def spruce(self, deleted_qubits):
        deleted_set = set(deleted_qubits)

        # Step 1: Remove any nn_pair or triangle term that includes a deleted qubit
        self.nn_pairs = {pair for pair in self.nn_pairs if not deleted_set.intersection(pair)}
        self.triangle_terms = {tri for tri in self.triangle_terms if not deleted_set.intersection(tri)}

        # Step 2: Create a new index mapping for the remaining qubits
        remaining_indices = sorted(set(self.index_map.keys()) - deleted_set)
        index_remap = {old_idx: new_idx for new_idx, old_idx in enumerate(remaining_indices)}

        # Step 3: Update site_map and index_map
        new_site_map = OrderedDict()
        new_index_map = {}

        for key, old_idx in self.site_map.items():
            if old_idx in index_remap:
                new_idx = index_remap[old_idx]
                new_site_map[key] = new_idx
                new_index_map[new_idx] = self.index_map[old_idx]

        self.site_map = new_site_map
        self.index_map = new_index_map
        self.num_sites = len(self.index_map)

        # Step 4: Remap nn_pairs and triangle_terms with new indices
        self.nn_pairs = {tuple(sorted((index_remap[i], index_remap[j]))) for i, j in self.nn_pairs}
        self.triangle_terms = {tuple(index_remap[i] for i in tri) for tri in self.triangle_terms}

    def __str__(self):
        return f"KagomeLattice with {self.num_sites} sites, {len(self.triangle_terms)} triangles"


from ncon_torch import ncon
import numpy as np
import copy
from scipy.linalg import svd, block_diag

###################################
##### Tensor methods ##############
###################################

default_msvr = 1e-8
def site_svd(site, which='left', msvr=None, max_dim=None, return_sv=False):
    """
    does an asymmetric SVD, keeping spin indice(s) on 'which' side, truncates depending on either max_dim 
    or min_sv_ratio
    """
    sh = site.shape
    is_mpo = len(sh) == 4
    
    if which == 'left': 
        if is_mpo:
            data = site.transpose(0,2,3,1).reshape(sh[0] * sh[2] * sh[3], sh[1])
        else: 
            data = site.transpose(0,2,1).reshape(sh[0] * sh[2], sh[1])
            
    else: 
        if is_mpo: 
            data = site.reshape(sh[0], sh[1] * sh[2] * sh[3])
        else: 
            data = site.reshape(sh[0], sh[1] * sh[2]) 
    
    u,s,vh = svd(data, full_matrices=False, lapack_driver='gesvd') 
    if msvr is not None: 
        s = s[s>msvr*s[0]]
    elif max_dim is not None:
        dim = min(max_dim, len(s[s>default_msvr*s[0]]))
        s = s[:dim]
        
    u = u[:,:len(s)] 
    vh = vh[:len(s),:]
    
    if which == 'left': 
        if is_mpo:
            left_tensor = u.reshape(sh[0], sh[2], sh[3], len(s)).transpose(0,3,1,2)
        else: 
            left_tensor = u.reshape(sh[0], sh[2], len(s)).transpose(0,2,1)
        right_tensor = np.diag(s) @ vh if not return_sv else vh
            
    else: 
        left_tensor = u @ np.diag(s) if not return_sv else u
        if is_mpo:
            right_tensor = vh.reshape(len(s), sh[1], sh[2], sh[3])
        else: 
            right_tensor = vh.reshape(len(s), sh[1], sh[2])
    
    if return_sv: 
        return left_tensor, right_tensor, s
    else:
        return left_tensor, right_tensor
    
def decompose_site(data): 
    """ 
    decomposes sites of an MPS or MPDO after action of a two-qubit gate 
    data.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    left_bond, right_bond)
    """
    sh = data.shape
    is_mpo = len(sh) == 6
    
    if is_mpo:
        data = data.transpose(0,1,4,2,3,5).reshape(sh[0]*sh[1]*sh[4], sh[2]*sh[3]*sh[5])
    else:
        data = data.transpose(0,2,1,3).reshape(sh[0]*sh[2], sh[1]*sh[3])
    
    u,s,vh = svd(data, full_matrices=False, lapack_driver='gesvd')    
    u, vh = u @ np.diag(np.sqrt(s)), np.diag(np.sqrt(s)) @ vh

    if is_mpo:
        top_site = u.reshape(sh[0],sh[1],sh[4],len(s)).transpose(2,3,0,1)
        bottom_site = vh.reshape(len(s),sh[2],sh[3],sh[5]).transpose(0,3,1,2)
    else:
        top_site = u.reshape(sh[0],sh[2],len(s)).transpose(1,2,0)
        bottom_site = vh.reshape(len(s),sh[1],sh[3]).transpose(0,2,1)
        
    return top_site, bottom_site

def decompose_three_site(data): 
    """ 
    decomposes sites of an MPDO after action of a three-site gate 
    data.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    third_top_spin, third_bottom_spin, left_bond, right_bond)
    """
    sh = data.shape
    data = data.transpose(0,1,2,4,3,5,6,7).reshape(sh[0],sh[1],sh[2]*sh[4],sh[3]*sh[5], sh[6], sh[7])
    
    top_site, bottom_site = decompose_site(data)
    bottom_site = bottom_site.reshape(bottom_site.shape[0], 
                                      sh[7], sh[2], sh[4], sh[3], sh[5]).transpose(2,4,3,5,0,1)
    middle_site, bottom_site = decompose_site(bottom_site)
    return top_site, middle_site, bottom_site 

def decompose_four_site(data): 
    """ 
    decomposes sites of an MPDO after action of a four-site gate 
    data.shape = (first_top_spin, first_bottom_spin, second_top_spin, second_bottom_spin, 
                    third_top_spin, third_bottom_spin, fourth_top_spin, fourth_bottom_spin, 
                    left_bond, right_bond)
    """
    sh = data.shape
    data = data.transpose(0,2,1,3,4,6,5,7,8,9).reshape(sh[0]*sh[2],sh[1]*sh[3],sh[4]*sh[6],sh[5]*sh[7], sh[8], sh[9])
    
    top_half, bottom_half = decompose_site(data)
    
    top_half = top_half.reshape(sh[8], top_half.shape[1], sh[0], sh[2], sh[1], sh[3]).transpose(2,4,3,5,0,1)
    bottom_half = bottom_half.reshape(bottom_half.shape[0], sh[9], sh[4], sh[6], sh[5], sh[7]).transpose(2,4,3,5,0,1)
    
    first_site, second_site = decompose_site(top_half)
    third_site, fourth_site = decompose_site(bottom_half)    
    return first_site, second_site, third_site, fourth_site 


############################################
############### CLASSES ####################
############################################

class TensorNetwork:
    def __init__(self, sites):
        self.sites = sites
        self.num_sites = len(sites)
        self.local_dim = sites[0].shape[-1]
        
    def __add__(self, other): 
        raise NotImplementedError("Subclasses must implement this method")
    
    def __mul__(self, scalar): 
        sites = copy.deepcopy(self.sites)
        sites[0] = sites[0] * scalar
        return type(self)(sites)
        
    def __matmul__(self, other):
        raise NotImplementedError("Subclasses must implement this method")
        
    def get_skeleton(self): 
        return [site.shape for site in self.sites]
    
    def get_max_dim(self):
        return max(max(shape) for shape in self.get_skeleton())
    
    def get_site_norms(self): 
        return [np.linalg.norm(np.ravel(site)) for site in self.sites]
    
    def conj(self): 
        sites = [copy.deepcopy(site).conj() for site in self.sites]
        return type(self)(sites)
        
class MPS(TensorNetwork): 
    """ a site is a np.array of shape (left_bond_dim, right_bond_dim, local_dim) """
    def __init__(self, sites):
        super().__init__(sites)
    
    def __matmul__(self, other):
        if isinstance(other, MPS):
            return self.overlap(other)
        else:
            raise TypeError("Unsupported operand type for @")
    
    def overlap(self, other, scaled=False):
        scale = np.sqrt(self.local_dim) if scaled else 1.0
        
        tensor = ncon((self.sites[0], other.sites[0].conj()), ([-1,-3,1],[-2,-4,1])) * scale
        for i in range(1,self.num_sites):
            tensor = ncon((tensor, self.sites[i]), ([-1,-2,1,-3],[1,-4,-5]))
            tensor = ncon((tensor, other.sites[i].conj()), ([-1,-2,1,-3,2],[1,-4,2])) * scale
        return np.trace(tensor.reshape(tensor.shape[0] * tensor.shape[1], tensor.shape[2] * tensor.shape[3]))

    def norm(self): 
        return self.overlap(self)
    
    def normalize(self): 
        n = self.norm()
        self.sites = [site/np.sqrt(np.abs(n))**(1/self.num_sites) for site in self.sites]
    
    def canonize(self, center): 
        """ orthogonalize self.state at center; 
        upto center-1 we have left-canonical; 
        from center+1 we have right-canonical """
        sites = copy.deepcopy(self.sites)
        if sites[0].shape[0] > 1:
            raise NotImplementedError("canonization for periodic boundaries has not been implemented")
        
        for i in range(self.num_sites-1, center-1, -1):
            left_tensor, right_tensor = site_svd(sites[i], 'right', None, None, False)
            sites[i] = right_tensor
            sites[i-1] = ncon((sites[i-1], left_tensor), ([-1,1,-3], [1,-2]))

        for i in range(0, center-1):
            left_tensor, right_tensor = site_svd(sites[i], 'left', None, None, False)
            sites[i] = left_tensor
            sites[i+1] = ncon((right_tensor, sites[i+1]), ([-1,1], [1,-2,-3]))

        left_tensor, right_tensor = site_svd(sites[center], 'left', None, None, False)
        sites[center] = left_tensor
        
        if center + 1 <= self.num_sites - 1:
            sites[center+1] = ncon((right_tensor, sites[center+1]), ([-1,1], [1,-2,-3]))
        else: 
            print("MPS has become normalized")       
        
        self.center = center
        self.sites = sites
        
    def compress(self, msvr=None, max_dim=None):
        sites = copy.deepcopy(self.sites)
        if sites[0].shape[0] > 1: 
            raise NotImplementedError("compression for periodic boundaries has not been implemented")

        # first we orthogonalize
        for i in range(self.num_sites-1, 0, -1):
            left_tensor, right_tensor = site_svd(sites[i], 'right', None, None)
            sites[i] = right_tensor
            sites[i-1] = ncon((sites[i-1], left_tensor), ([-1,1,-3], [1,-2]))
        
        # then we sweep down and truncate
        for i in range(0, self.num_sites-1):
            left_tensor, right_tensor = site_svd(sites[i], 'left', msvr, max_dim)
            sites[i] = left_tensor
            sites[i+1] = ncon((right_tensor, sites[i+1]), ([-1,1], [1,-2,-3]))
        
        # then we sweep up and truncate
        for i in range(self.num_sites-1, 0, -1):
            left_tensor, right_tensor = site_svd(sites[i], 'right', msvr, max_dim)
            sites[i] = right_tensor
            sites[i-1] = ncon((sites[i-1], left_tensor), ([-1,1,-3], [1,-2]))

        return MPS(sites)
    
    def to_vector(self):
        return contract_sites(self.sites)[0,0,:].reshape(self.local_dim**self.num_sites)
    
    
    def evaluate(self, basis): 
        return 0.0
 
class MPO(TensorNetwork): 
    """ a site is a np.array of shape (left_bond_dim, right_bond_dim, top_local_dim, bottom_local_dim) """
    def __init__(self, sites):
        super().__init__(sites)
        
    def __add__(self, other): 
        new_sites = []
        
        for i, (s,o) in enumerate(zip(self.sites, other.sites)):
            if i == 0: 
                site = np.zeros((1, s.shape[1]+o.shape[1], self.local_dim, self.local_dim), dtype=complex)
                for i in range(self.local_dim):
                    for j in range(self.local_dim): 
                        site[0,:,i,j] = np.hstack((s[0,:,i,j], o[0,:,i,j]))
                        
            elif i == self.num_sites - 1: 
                site = np.zeros((s.shape[0]+o.shape[0], 1, self.local_dim, self.local_dim), dtype=complex)
                for i in range(self.local_dim):
                    for j in range(self.local_dim): 
                        site[:,0,i,j] = np.hstack((s[:,0,i,j], o[:,0,i,j]))
            
            else: 
                site = np.zeros((s.shape[0]+o.shape[0], s.shape[1]+o.shape[1], self.local_dim, self.local_dim), dtype=complex)
                for i in range(self.local_dim):
                    for j in range(self.local_dim): 
                        site[:,:,i,j] = block_diag(s[:,:,i,j], o[:,:,i,j])

            new_sites.append(site)
        return type(self)(new_sites)
    
    def __matmul__(self, other):
        if isinstance(other, MPS):
            return self.act(other)
        elif isinstance(other, MPO):
            return self.compose(other)
        else:
            raise TypeError("Unsupported operand type for @")
            
    def act(self, mps):
        new_sites = []
        for site_a, site_b in zip(self.sites, mps.sites):
            ldim = site_a.shape[0] * site_b.shape[0]
            rdim = site_a.shape[1] * site_b.shape[1]
            new_site = ncon((site_a, site_b), ([-1,-3,-5,1],[-2,-4,1])).reshape(ldim, rdim, mps.local_dim)
            new_sites.append(new_site)
        return MPS(new_sites)
    
    def compose(self, mpo):
        new_sites = []
        for site_a, site_b in zip(self.sites, mpo.sites):
            ldim = site_a.shape[0] * site_b.shape[0]
            rdim = site_a.shape[1] * site_b.shape[1]
            sdim = self.local_dim
            new_site = ncon((site_a, site_b), ([-1,-3,-5,1],[-2,-4,1,-6])).reshape(ldim, rdim, sdim, sdim)
            new_sites.append(new_site)
        return type(self)(new_sites)
    
    def overlap(self, other, scaled=True): 
        scale = self.local_dim if scaled else 1.0
        tensor = ncon((self.sites[0], other.sites[0]), ([-1,-3,1,2],[-2,-4,2,1]))/scale
        for i in range(1,self.num_sites):
            tensor = ncon((tensor, self.sites[i]), ([-1,-2,1,-3],[1,-4,-5,-6]))
            tensor = ncon((tensor, other.sites[i]), ([-1,-2,1,-3,2,3],[1,-4,3,2]))/scale
        return np.trace(tensor.reshape(tensor.shape[0] * tensor.shape[1], tensor.shape[2] * tensor.shape[3]))
    
    def trace(self, scaled=True): 
        scale = self.local_dim if scaled else 1.0
        if self.num_sites == 1:
            return np.trace(self.sites[0])
        else: 
            tr = ncon((self.sites[0]), ([-1,-2,1,1]))/scale
            for i in range(1,self.num_sites): 
                tr = ncon((tr, self.sites[i]), ([-1,1],[1,-2,2,2]))/scale
            return np.trace(tr)
    
    def to_matrix(self): 
        tensor = self.sites[0]
        for i in range(1,self.num_sites):
            left_dim = tensor.shape[0]
            right_dim = self.sites[i].shape[1]
            local_dim = self.local_dim**(i+1)
            tensor = ncon((tensor, self.sites[i]), ([-1,1,-3,-5],[1,-2,-4,-6])).reshape(left_dim,right_dim,local_dim,local_dim)
        return ncon((tensor), (1,1,-1,-2))
    
    def compress(self, msvr=None, max_dim=None):
        sites = copy.deepcopy(self.sites)
        if sites[0].shape[0] > 1: 
            raise NotImplementedError("compression for periodic boundaries has not been implemented")

        # first we orthogonalize
        for i in range(self.num_sites-1, 0, -1):
            left_tensor, right_tensor = site_svd(sites[i], 'right', None, None)
            sites[i] = right_tensor
            sites[i-1] = ncon((sites[i-1], left_tensor), ([-1,1,-3,-4], [1,-2]))
        
        # then we sweep down and truncate
        for i in range(0, self.num_sites-1):
            left_tensor, right_tensor = site_svd(sites[i], 'left', msvr, max_dim)
            sites[i] = left_tensor
            sites[i+1] = ncon((right_tensor, sites[i+1]), ([-1,1], [1,-2,-3,-4]))
        
        # then we sweep up and truncate
        for i in range(self.num_sites-1, 0, -1):
            left_tensor, right_tensor = site_svd(sites[i], 'right', msvr, max_dim)
            sites[i] = right_tensor
            sites[i-1] = ncon((sites[i-1], left_tensor), ([-1,1,-3,-4], [1,-2]))

        return type(self)(sites)
    
class MPDO(MPO):
    def __init__(self, sites):
        super().__init__(sites)
    
    def partial_trace(self, indices): 
        sites = copy.deepcopy(self.sites)
        offset = 0

        for i in indices:
            cur_idx = i-offset
            sites[cur_idx] = ncon((sites[cur_idx]), ([-1,-2,1,1]))

            if cur_idx ==  len(sites)-1: 
                sites[cur_idx-1] = ncon((sites[cur_idx-1], sites[cur_idx]), ([-1,1,-3,-4],[1,-2]))            
            else: 
                sites[cur_idx+1] = ncon((sites[cur_idx], sites[cur_idx+1]), ([-1,1],[1,-2,-3,-4])) 

            sites.pop(cur_idx)
            offset += 1

        return MPDO(sites)
    
    def apply_isometry(self, site_indices, isometries): 
        """ contracts with isometries """
        site_indices = [site_indices] if type(site_indices) == int else site_indices
        isometries = [isometries] * len(site_indices) if type(isometries) == np.ndarray else isometries
        sites = copy.deepcopy(self.sites)
        
        for site_idx, isometry in zip(site_indices, isometries): 
            sites[site_idx] = ncon((sites[site_idx], isometry, isometry.conj()), ([-1,-2,1,2],[1,-3],[2,-4]))
        
        return MPDO(sites)                                         
    
    def project(self, site_idx, vector): 
        """ contracts with a projector (just a flat isometry) """
        sites = copy.deepcopy(self.sites)
        virt_site = sites.pop(site_idx)
        virt_site = ncon((virt_site, vector, vector.conj()), ([-1,-2,1,2],[1],[2]))

        if self.num_sites == 1: 
            sites = [virt_site]
        elif site_idx == self.num_sites - 1: 
            sites[-1] = ncon((sites[-1], virt_site), ([-1,1,-3,-4],[1,-2]))
        else: 
            sites[site_idx] = ncon((virt_site, sites[site_idx]), ([-1,1],[1,-2,-3,-4]))
        
        return MPDO(sites)
    
    def outcome_prob(self, site_indices, outcomes):
        if type(outcomes) == str: 
            vectors = [] 
            for outcome in outcomes: 
                v = np.zeros(self.local_dim)
                v[int(outcome)] = 1.0
                vectors.append(v)   
        else: 
            vectors = outcomes
            
        offset = 0 
        proj_mpo = copy.deepcopy(self)
        for site_idx, vector in zip(site_indices, vectors):
            site_idx = site_idx - offset
            proj_mpo = proj_mpo.project(site_idx, vector)
            offset += 1
        
        return proj_mpo.trace(scaled=False).real
            
    def diag(self): 
        """ returns MPDO representing diagonal part of the operator """
        sites = copy.deepcopy(self.sites)
        for k in range(self.num_sites):
            site = sites[k]
            s1,s2 = site.shape[0], site.shape[1]
            for i in range(s1):
                for j in range(s2):
                    site[i,j,:,:] = np.diag(np.diag(site[i,j,:,:]))
            sites[k] = site
        return MPDO(sites)
    
    def conj(self): 
        sites = [copy.deepcopy(site).conj().transpose(0,1,3,2) for site in self.sites]
        return type(self)(sites)
    
def contract_sites(site_list): 
    """ takes a list of MPS or MPDO sites and contracts intermediate bonds
    for MPDO sites the two local dimensions are combined into a super local_dim  
    final shape  = (bond_dim, bond_dim, total local_dim) """

    if len(site_list[0].shape) == 4: 
        site_list = [site.reshape(site.shape[0], site.shape[1], site.shape[2] * site.shape[3]) for site in site_list]

    site = site_list[0]
    for i in range(1, len(site_list)): 
        site = ncon((site, site_list[i]), ([-1,1,-3], [1,-2,-4])).reshape(
            site.shape[0], site_list[i].shape[1], site.shape[2]*site_list[i].shape[2])
        
    return site

def decompose_site_n(data, msvr=None, max_dim=None): 
    """ 
    decomposes sites of an MPS or MPDO by combining left_bond and top_spin, and right_bond and bottom_spin
    data.shape = (left_bond, right_bond, top_spin, bottom_spin)
    top_site.shape = (left_bond, new_bond, top_spin)
    bottom_site.shape = (new_bond, right_bond, bottom_spin)
    """
    sh = data.shape
    data = data.transpose(0, 2, 1, 3).reshape(sh[0] * sh[2], sh[1] * sh[3])  
    # shape has become (left_bond * top_spin, right_bond * bottom_spin)

    u,s,vh = svd(data, full_matrices=False, lapack_driver='gesvd')    
    u, vh = u @ np.diag(np.sqrt(s)), np.diag(np.sqrt(s)) @ vh

    if msvr is not None: 
        s = s[s>msvr*s[0]]
    elif max_dim is not None:
        dim = min(max_dim, len(s[s>default_msvr*s[0]]))
        s = s[:dim]
    else: 
        s = s[s>default_msvr*s[0]]

    u = u[:,:len(s)] 
    vh = vh[:len(s),:]

    top_site = u.reshape(sh[0],sh[2],len(s)).transpose(0,2,1)
    bottom_site = vh.reshape(len(s),sh[1],sh[3])
        
    return top_site, bottom_site

def decimate(con_site, local_dim): 
    """ decimates sites of an MPS or MPDO after the action of span-local gate 
    con_site.shape = (bond_dim, bond_dim, local_dim**span) """
    span = int(np.log(con_site.shape[2]) / np.log(local_dim))

    dec_sites = []
    for i in range(span-1): 
        con_site = con_site.reshape(con_site.shape[0], con_site.shape[1], local_dim, local_dim**(span-i-1))
        # shape has become (b, b, d, d*d*d*...*d)
        top, con_site = decompose_site_n(con_site)
        dec_sites.append(top)

    dec_sites.append(con_site)
    return dec_sites

def transfer_matrices(mps1, mps2): 
    tm_list = []
    for site1, site2 in zip(mps1.sites, mps2.sites):
        tm = ncon((site1, site2.conj()), ([-1,-3,1],[-2,-4,1]))
        tm_list.append(tm)
    return tm_list


from tqdm import tqdm

class BaseSim:
    """ 
    accepts Circuit object and Noise object 
    different simulators use different representations of quantum state
    and different methods for applying gates and measurements
    """

    def __init__(self, circ, init_state=None, nm=None):
        self.circ = circ.copy()
        self.num_sites = circ.num_sites
        self.local_dim = circ.local_dim
        self.nm = nm
        self.init_state = init_state
        self.state = None
        self.ran = False

    def initialize_state(self):
        raise NotImplementedError("must be implemented by child classes.")

    def apply_gate(self, gate):
        raise NotImplementedError("must be implemented by child classes.")

    def run(self, progress_bar=True, debug=0):
        """
        progress_bar: If True, display a progress bar during simulation.
        debug: If 1, prints intermediate state after each layer.
        debug: If 2, prints intermediate state after each gate.
        """
        self.initialize_state()
        circ = self.circ.build(self.nm) if not self.circ.built else self.circ
        layers = tqdm(circ.layers) if progress_bar else circ.layers

        for layer in layers:
            for gate in layer:
                self.apply_gate(gate)
                if debug == 2:
                    print(self)
            
            if debug == 1:
                print(self)

        if circ.meas_sites is not None:
            self.measure(circ.meas_sites)

        self.ran = True
        return self.state
    
    def measure(self, meas_sites):
        raise NotImplementedError("must be implemented by child classes.")

    def __str__(self):
        """ pretty prints the state """
        raise NotImplementedError("must be implemented by child classes.")


from .gates import *
from .circuits import Circuit
from .lattices import LinearLattice, KagomeLattice, ToricLattice
from .paulis import pauli_mapping
from scipy.linalg import expm

def terms_to_matrix(terms, num_sites, dense=True):
    """ builds full 2**n by 2**n matrix from a list of terms, where each term is a tuple (coefficient, active_sites, local_ops) """
    H = sp.csr_matrix((2**num_sites, 2**num_sites), dtype=complex)
    for coeff, active_sites, local_ops in terms:
        H += coeff * embed_operator(num_sites, active_sites, [pauli_mapping[op] for op in local_ops], factor=True)
    if dense: 
        return H.toarray()
    return H

class Hamiltonian:
    def __init__(self, coupling_types, coupling_strengths, field_types, field_strengths, lattice=None, num_x=None, locality='nn', periodic_x=False):
        """ 
        coupling_types is a list like ['xx', 'yx', 'zz'] 
        coupling_strengths is same or double the length of coupling_types
        field_types is a subset of ['x', 'y', 'z']
        field_strengths is same length as field_types
        locality is one of ['nn', 'nnn', 'a2a', 'long-range'] 
        """
        if lattice is not None:
            self.lattice = lattice
        elif num_x is not None:
            self.lattice = LinearLattice(num_x, periodic_x)
        else:
            raise ValueError("Either 'lattice' or 'num_x' must be provided.")

        self.num_sites = self.lattice.num_sites
        self.op_grouping = None
        self.coupling_types = coupling_types
        self.coupling_strengths = coupling_strengths
        self.field_types = field_types
        self.field_strengths = field_strengths
        self.locality = locality
        self.terms = []  # stores (coefficient, active_sites, local_ops)
        
        if locality == 'nn':
            for ops, strength in zip(coupling_types, coupling_strengths):
                for pairs in self.lattice.get_connections("nn"):
                    self.terms.append((strength, pairs, [ops[0], ops[1]]))

        elif locality == 'nnn':
            mid = len(coupling_strengths) // 2
            nn_strengths = coupling_strengths[:mid]
            nnn_strengths = coupling_strengths[mid:]
            nn_pairs = self.lattice.get_connections("nn")
            nnn_pairs = self.lattice.get_connections("nnn")

            for ops, nn_strength, nnn_strength in zip(coupling_types, nn_strengths, nnn_strengths):
                for pairs in nn_pairs:
                    self.terms.append((nn_strength, pairs, [ops[0], ops[1]]))
                for pairs in nnn_pairs:
                    self.terms.append((nnn_strength, pairs, [ops[0], ops[1]]))

        elif locality == 'a2a': 
            for ops, strength in zip(coupling_types, coupling_strengths):
                for pairs in self.lattice.get_connections("a2a"):
                    self.terms.append((strength, pairs, [ops[0], ops[1]]))

        elif locality == 'long-range': 
            for ops, strength in zip(coupling_types, coupling_strengths):
                for pairs in self.lattice.get_connections("a2a"):
                    self.terms.append((strength/np.abs(pairs[0]-pairs[1]), pairs, [ops[0], ops[1]]))

        for op, strength in zip(field_types, field_strengths):
            if np.abs(strength) > 1e-10:  # avoid adding zero fields
                for i in range(self.num_sites):
                    self.terms.append((strength, [i], [op]))

    def matrix(self, dense=False):
        return terms_to_matrix(self.terms, self.num_sites, dense=dense)

    def evals(self):
        return np.linalg.eigvalsh(self.matrix(dense=True)) 

    def ground(self): 
        H = self.matrix()
        return sp.linalg.eigsh(H, k=1, which='SA')       

    def propagator(self, time, dense=False):
        H = self.matrix()
        if dense:
            return sp.linalg.expm(-1j * time * H).toarray()
        return sp.linalg.expm(-1j * time * H)
    
    def evolve(self, time, init_state, dense=False):
        final_state = sp.linalg.expm_multiply(-1j * time * self.matrix(), init_state)
        if dense:
            return final_state.toarray()[:,0]
        return final_state

    def trotter_circ(self, step_size, num_steps, order=1):
        """ 
        returns Circuit object for Trotter decomposition of propagator
        currently only works if locality is nn and Hamiltonian terms can be divided into two mutually commuting groups saved in op_grouping
        so XY or Heisenberg with NNN and/or with field terms will not work
        """
        assert self.op_grouping is not None, "op_grouping is not defined or is invalid"
        gate_list = []
        if order == 1: 
            layer_list = [self.layer_from_group(group, step_size, self.num_sites) for group in self.op_grouping]
            for i in range(num_steps):
                for layer in layer_list:
                    gate_list += layer

        elif order == 2:
            l0h = self.layer_from_group(self.op_grouping[0], step_size/2, self.num_sites)
            l0 = self.layer_from_group(self.op_grouping[0], step_size, self.num_sites)
            l1 = self.layer_from_group(self.op_grouping[1], step_size, self.num_sites)
            
            gate_list += l0h
            for i in range(num_steps-1):
                gate_list += l1
                gate_list += l0

            gate_list += l1
            gate_list += l0h

        else: 
            raise NotImplementedError("Only first and second order Trotter decomposition is supported")
        
        return Circuit(gate_list, self.num_sites)

    def layer_from_group(self, group, time, num_sites):
        """ group is something like 'xx' or 'z' or 'xy' or 'even' or 'odd' """
        if group in ['even', 'odd']:
            generator_mat = sum([c*embed_operator(2, [0, 1], [pauli_mapping[t[0]], pauli_mapping[t[1]]]) for c,t in zip(self.coupling_strengths, self.coupling_types)])
        else: 
            c = self.coupling_strengths[self.coupling_types.index(group)] if len(group) == 2 else self.field_strengths[self.field_types.index(group)]
            generator_mat = c*embed_operator(len(group), [i for i in range(len(group))], [pauli_mapping[op] for op in group])

        gate_mat = expm(-1j * time * generator_mat.toarray())
        gate_list = []

        if len(group) == 1: 
            for i in range(num_sites): 
                gate = Gate('R1', [i], gate_mat)
                gate_list.append(gate)
        
        elif len(group) == 2: 
            for i in range(0, num_sites, 2):
                gate = Gate('R2', [i, i + 1], gate_mat)
                gate_list.append(gate)
            for i in range(1, num_sites-1, 2):
                gate = Gate('R2', [i, i + 1], gate_mat)
                gate_list.append(gate)

        elif group == 'even':
            for i in range(0, num_sites, 2):
                gate = Gate('R2', [i, i + 1], gate_mat)
                gate_list.append(gate)

        elif group == 'odd':
            for i in range(1, num_sites-1, 2):
                gate = Gate('R2', [i, i + 1], gate_mat)
                gate_list.append(gate)

        else: 
            raise NotImplementedError("Only field and nearest-neighbor couplings are supported")

        return gate_list
    
    def __str__(self): 
        return f"Hamiltonian with {self.num_sites} sites and {self.locality} {self.coupling_types} coupling(s) and {self.field_types} field"

class Heisenberg(Hamiltonian):
    def __init__(self, num_x, jx=1, jy=1, jz=1, h=0):
        super().__init__(
            coupling_types=['xx', 'yy', 'zz'],
            coupling_strengths=[jx, jy, jz],
            field_types=['z'],
            field_strengths=[h],
            num_x=num_x,
            locality='nn'
        )
        self.op_grouping = ['even', 'odd']

class HeisenbergNNN(Hamiltonian):
    def __init__(self, num_x, jx=1, jy=1, jz=1, jx2=1e-2, jy2=1e-2, jz2=1e-2, h=0):
        super().__init__(
            coupling_types=['xx', 'yy', 'zz'],
            coupling_strengths=[jx, jy, jz, jx2, jy2, jz2],
            field_types=['z'],
            field_strengths=[h],
            num_x=num_x,
            locality='nnn'
        )

class XY(Hamiltonian):
    def __init__(self, num_x, j=1, h=0):
        super().__init__(
            coupling_types=['xx', 'yy'],
            coupling_strengths=[j, j],
            field_types=['z'],
            field_strengths=[h],
            num_x=num_x,
            locality='nn'
        )
        self.op_grouping = ['even', 'odd']

class XYNNN(Hamiltonian):
    def __init__(self, num_x, j=1, j2=0.01, h=0):
        super().__init__(
            coupling_types=['xx', 'yy'],
            coupling_strengths=[j, j, j2, j2],
            field_types=['z'],
            field_strengths=[h],
            num_x=num_x,
            locality='nnn'
        )

    def layer_from_group(self, group, time, num_sites):
        if group in ['even', 'odd']:
            generator_mat = sum(
                c * embed_operator(2, [0, 1], [pauli_mapping[t[0]], pauli_mapping[t[1]]])
                for c, t in zip(self.coupling_strengths, self.coupling_types)
            )
        else:
            generator_mat = sum(
                c * embed_operator(3, [0, 1, 2], [pauli_mapping[t[0]], pauli_mapping[t[1]], pauli_mapping[t[2]]])
                for c, t in zip(self.coupling_strengths, self.coupling_types)
            )

class TFI(Hamiltonian):
    def __init__(self, num_x, jz=1, h=0, periodic_x=False):
        super().__init__(
            coupling_types=['zz'],
            coupling_strengths=[jz],
            field_types=['x'],
            field_strengths=[h],
            num_x=num_x,
            locality='nn', 
            periodic_x=periodic_x
        )
        self.op_grouping = ['x', 'zz']

class LMG(Hamiltonian):
    def __init__(self, num_x, jx=1, h=0):
        super().__init__(
            coupling_types=['xx'],
            coupling_strengths=[jx],
            field_types=['z'],
            field_strengths=[h],
            num_x=num_x,
            locality='a2a'
        )

class ChiralHeisenberg(Hamiltonian):
    def __init__(self, row_layout, jx=1, jy=1, jz=1, chiral_strength=0.01, h=0, sp=None):
        lattice = KagomeLattice(row_layout)
        if sp is not None:
            lattice.spruce(sp)
        super().__init__(coupling_types=['xx', 'yy', 'zz'],
                         coupling_strengths=[jx, jy, jz],
                         field_types=['z'],
                         field_strengths=[h],
                         lattice=lattice,
                         locality='nn')
        
        self.chiral_strength = chiral_strength
        if np.abs(chiral_strength) > 1e-10:
            for triplet in self.lattice.triangle_terms:
                self.terms.extend(self.chiral_terms(triplet))
       
    def pairwise_terms(self, pairs):
        return [(j, p, op) for p in pairs for (j,op) in zip(self.coupling_strengths, self.coupling_types)]

    def chiral_terms(self, triplet):
        ops = ['xyz', 'yzx', 'zxy', 'xzy', 'zyx', 'yxz']
        return [(self.chiral_strength if op in ['xyz', 'yzx', 'zxy'] else -self.chiral_strength, triplet, op) for op in ops]

    def primitive_matrices(self):
        """ returns small 8 by 8 matrix for triangle primitives and 4 by 4 matrix for edge primitives """
        triangle_terms = self.pairwise_terms([(0,1), (1,2), (0,2)]) + self.chiral_terms((0,1,2))
        tmat = sum([coeff * embed_operator(3, sites, [pauli_mapping[op] for op in local_ops], factor=True) for coeff, sites, local_ops in triangle_terms])

        edge_terms = self.pairwise_terms([(0,1),])
        emat = sum([coeff * embed_operator(2, sites, [pauli_mapping[op] for op in local_ops], factor=True) for coeff, sites, local_ops in edge_terms])
        return tmat.toarray(), emat.toarray()

    def grouped_terms(self): 
        """ returns a list of (indices, mat) where mat acts on qubits at indices, not mat is either 4 by 4 or 8 by 8 and each index triplet or pair only appears once 
        used by ham_action function; output cannot be used to compute the full Hamiltonian matrix since indices are possibly non-contiguous """
        tmat, emat = self.primitive_matrices()
        triangle_edges = [set(pair) for t in self.lattice.triangle_terms for pair in [(t[0], t[1]), (t[1], t[2]), (t[0], t[2])]]
        edge_terms = [pair for pair in self.lattice.nn_pairs if set(pair) not in triangle_edges]
        return [(indices, tmat) for indices in self.lattice.triangle_terms] + [(indices, emat) for indices in edge_terms]
            
class ToricCode(Hamiltonian):
    def __init__(self, Lx, Ly, hx=1.0, hz=1.0):
        self.lattice = ToricLattice(Lx, Ly)
        self.num_sites = self.lattice.num_sites
        self.terms = []

        for v_edges in self.lattice.vertex_terms:
            self.terms.append((-hx, v_edges, ['x'] * 4))

        for p_edges in self.lattice.plaquette_terms:
            self.terms.append((-hz, p_edges, ['z'] * 4))

    def __str__(self):
        return (f"Toric Code Hamiltonian on {self.lattice.Lx}×{self.lattice.Ly} "
                f"lattice with {self.num_sites} qubits")


import numpy as np
import itertools

pauli_I = np.array([[1,0],[0,1]], dtype=complex)
pauli_X = np.array([[0,1],[1,0]], dtype=complex)
pauli_Y = np.array([[0,-1.j],[1.j,0]], dtype=complex)
pauli_Z = np.array([[1,0],[0,-1]], dtype=complex)
pauli_mapping = {'i': pauli_I, 'x': pauli_X, 'y': pauli_Y, 'z': pauli_Z}

def pauli_commute(pauli1, pauli2):
    anticommute_count = 0
    for p1, p2 in zip(pauli1, pauli2):
        if p1 == 'i' or p2 == 'i':
            continue  
        if p1 != p2:
            anticommute_count += 1
    return anticommute_count % 2 == 0

def pauli_multiply(pauli1, pauli2):
    pauli_dict = {'i': 0, 'x': 1, 'y': 2, 'z': 3}
    inverse_pauli_dict = {0: 'i', 1: 'x', 2: 'y', 3: 'z'}

    epsilon = np.zeros((4, 4, 4), dtype=int)
    epsilon[1, 2, 3] = epsilon[2, 3, 1] = epsilon[3, 1, 2] = 1
    epsilon[3, 2, 1] = epsilon[1, 3, 2] = epsilon[2, 1, 3] = -1

    result_pauli = []
    phase = 1

    for p1, p2 in zip(pauli1, pauli2):
        i, j = pauli_dict[p1], pauli_dict[p2]

        if i == 0: 
            result_pauli.append(p2)
        elif j == 0:
            result_pauli.append(p1)
        elif i == j:  
            result_pauli.append('i')
        else:
            k = 6 - (i + j)  # Ensures X(1), Y(2), Z(3) → 1+2+3 = 6, so k = the missing index
            sign = epsilon[i, j, k]  # Get the correct sign
            result_pauli.append(inverse_pauli_dict[k])

            phase *= 1j * sign

    return "".join(result_pauli), phase

def pauli_strings(n):
    pauli_ops = 'ixyz'
    return [''.join(p) for p in itertools.product(pauli_ops, repeat=n)]

def random_pauli_string(n):
    chars = 'ixyz'
    out = ''
    for i in range(n):
        out += chars[np.random.choice(len(chars))]
    return out


from .gates import Gate, SX, SX01
from .paulis import pauli_X, pauli_Y, pauli_Z
import numpy as np
import scipy
import torch

xxN, yyN, zzN = np.kron(pauli_X, pauli_X), np.kron(pauli_Y, pauli_Y), np.kron(pauli_Z, pauli_Z)
pauli_XT = torch.tensor([[0,1],[1,0]], dtype=torch.complex128)
pauli_YT = torch.tensor([[0,-1.j],[1.j,0]], dtype=torch.complex128)
pauli_ZT = torch.tensor([[1,0],[0,-1]], dtype=torch.complex128)
xxT = torch.kron(pauli_XT, pauli_XT)
yyT = torch.kron(pauli_YT, pauli_YT)
zzT = torch.kron(pauli_ZT, pauli_ZT)

def arr_constructor(x, backend="numpy"):
    if backend == "torch":
        return torch.stack([torch.stack([elem for elem in row]) for row in x]).to(dtype=torch.complex128)
    else:
        return np.array(x, dtype=np.complex128)

class ParamGate(Gate):
    def __init__(self, name, indices, *args): 
        super().__init__(name, indices, None) 
        self.backend = "torch" if torch.is_tensor(args[0]) else "numpy"
        
        if type(args[0]) == np.ndarray: 
            mat = args[0]
            self.update_matrix(mat)
            self.angles = self.solve_angles()
        
        else:
            self.angles = [args] if type(args) == float else args
            mat = self.construct_matrix()
            self.update_matrix(mat)
            
    def __str__(self):
        return f"{self.name} gate on site(s) {self.indices} with angle(s) {self.angles}"

    def update_matrix(self, mat): 
        self.matrix = mat
        
    def construct_matrix(self): 
        raise NotImplementedError("Subclasses must implement this method")
        
    def solve_angles(self): 
        return "unsolved" # if subclass doesn't bother implementing we probably don't need angles
    
class RZ(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ", indices, *args)
    
    def construct_matrix(self): 
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[1,0],[0,xp.exp(1.j*self.angles[0])]], backend=self.backend)
    
class RZZ(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("RZZ", indices, *args)
    
    def construct_matrix(self): 
        expm = scipy.linalg.expm if self.backend == "numpy" else torch.linalg.matrix_exp
        zz = zzN if self.backend == "numpy" else zzT
        return arr_constructor(expm(-1.j*self.angles[0] * zz))
    
class CPhase(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("CPhase", indices, *args)
    
    def construct_matrix(self): 
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,xp.exp(1.j*self.angles[0])]], 
                               backend=self.backend)
    
    def solve_angles(self): 
        theta = np.angle(self.matrix[3,3])
        return (theta,)

class U(ParamGate):
    """ create this gate with the following signature: 
        U(indices, theta, phi, lambda) """
    def __init__(self, indices, *args):
        super().__init__("U", indices, *args)
    
    def construct_matrix(self): 
        t,p,l = self.angles
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[xp.cos(t/2), -xp.exp(1j*l)*xp.sin(t/2)],
                           [xp.exp(1j*p)*xp.sin(t/2), xp.exp(1j*(p+l))*xp.cos(t/2)]], 
                           backend=self.backend)
    
    def solve_angles(self): 
        phase = np.angle(self.matrix[0,0]) 
        self.matrix = self.matrix / np.exp(1.j*phase) # removes phase for top left entry
        t = np.arccos(self.matrix[0,0])*2
        if np.abs(np.sin(t/2)) > 1e-12:
            l = np.angle(-self.matrix[0,1]/np.sin(t/2))
            p = np.angle(self.matrix[1,1]/self.matrix[0,0])-l
        else: 
            l_plus_p = np.angle(self.matrix[1,1])
            l,p = l_plus_p/2, l_plus_p/2
        
        return t,p,l        
        
    def decompose(self, basis='ZSX'): 
        if basis == 'ZSX': 
            t,p,l = self.solve_angles()#self.angles
            i = self.indices
            
            return list(reversed([
                RZ(i,p+np.pi),
                SX(i),
                RZ(i,t+np.pi),
                SX(i), 
                RZ(i,l)
            ]))   # If A = BC, we must apply C first and then B
            
        else: 
            raise NotImplementedError(f"{basis} basis decomposition has not been implemented")
        
    def shallow_copy(self):
        return U(self.indices, *self.angles)

class Givens(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("Givens", indices, *args)
    
    def construct_matrix(self): 
        theta = self.angles[0]
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[1,0,0,0],
                                [0,xp.cos(theta/2),-xp.sin(theta/2),0],
                                [0,xp.sin(theta/2),xp.cos(theta/2),0],
                                [0,0,0,1]], 
                                backend=self.backend)
     
    def solve_angles(self): 
        theta = np.arccos(self.matrix[1,1])*2
        return (theta,)

class XYCoupling(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("XYCoupling", indices, *args)
    
    def construct_matrix(self): 
        theta = self.angles[0]
        expm = scipy.linalg.expm if self.backend == "numpy" else torch.linalg.matrix_exp
        xx = xxN if self.backend == "numpy" else xxT
        yy = yyN if self.backend == "numpy" else yyT
        return arr_constructor(expm(-1.j*theta * (xx + yy)), backend=self.backend)

class RZ01(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ01", indices, *args)
    
    def construct_matrix(self): 
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[1,0,0],[0,xp.exp(1.j*self.angles[0]),0],[0,0,1]], 
                               backend=self.backend)
    
class RZ12(ParamGate): 
    def __init__(self, indices, *args):
        super().__init__("RZ12", indices, *args)
    
    def construct_matrix(self): 
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[1,0,0],[0,1,0],[0,0,xp.exp(1.j*self.angles[0])]], 
                               backend=self.backend)
    
class U01(ParamGate):
    def __init__(self, indices, *args):
        super().__init__("U01", indices, *args)
    
    def construct_matrix(self): 
        t,p,l = self.angles
        xp = np if self.backend == "numpy" else torch
        return arr_constructor([[xp.cos(t/2), -xp.exp(1j*l)*xp.sin(t/2), 0],
                           [xp.exp(1j*p)*xp.sin(t/2), xp.exp(1j*(p+l))*xp.cos(t/2), 0],
                           [0, 0, 1]], 
                           backend=self.backend)
    
    def solve_angles(self): 
        return None # TODO implement solver 
    
    def decompose(self, basis='ZSX'): 
        if basis == 'ZSX': 
            t,p,l = self.angles
            i = self.indices
            return list(reversed([
                RZ01(i,p+np.pi),
                SX01(i),
                RZ01(i,t+np.pi),
                SX01(i), 
                RZ01(i,l)
            ]))
        else: 
            raise NotImplementedError(f"{basis} basis decomposition has not been implemented")

def kak_unitary(params):
    backend = "torch" if torch.is_tensor(params) else "numpy"
    xp = np if backend == "numpy" else torch
    xx = xxN if backend == "numpy" else xxT
    yy = yyN if backend == "numpy" else yyT
    zz = zzN if backend == "numpy" else zzT
    expm = scipy.linalg.expm if backend == "numpy" else torch.linalg.matrix_exp

    left_mat = xp.kron(U(0, *params[0:3]).matrix, U(0, *params[3:6]).matrix)
    right_mat = xp.kron(U(0, *params[9:12]).matrix, U(0, *params[12:15]).matrix)    
    arg = sum([p*P for p,P in zip(params[6:], [xx, yy, zz])])
    center_mat = expm(1.j* arg)
    
    return left_mat @ center_mat @ right_mat


from setuptools import setup, find_packages

setup(
    name="qaravan",
    packages=find_packages(where="src"),
    python_requires='>=3.8',
    package_dir={"": "src"},
    version="0.1.56", 
    author="Faisal Alam",
    author_email="mfalam2@illinois.edu",
    description="Unified classical simulation of noiseless and noisy quantum circuits",
    install_requires=[
    "numpy>=1.20,<2.0",
    "scipy>=1.5.2,<2.0",
    "sympy >= 1.6",
    "matplotlib>=3.7,<4.0",
    "tqdm>=4.67.1",
    "ncon_torch>=0.3",
    "torch>=1.12",
    "stim>=1.14.0"
]
)



from .gates import ID, SuperOp, random_unitary, Gate
from .param_gates import kak_unitary
from .noise import gate_time, ThermalNoise
import numpy as np
import copy
from functools import reduce 

class Circuit:
    def __init__(self, gate_list, n=None, local_dim=2, meas_sites=None):
        self.gate_list = gate_list

        if len(gate_list) != 0: 
            sites = np.concatenate([gate.indices for gate in gate_list])
            self.num_sites = int(max(sites) + 1) if n is None else n
            self.local_dim = gate_list[0].get_local_dim()

        else: 
            self.num_sites = n 
            self.local_dim = local_dim

        self.layers = None
        self.built = False
        self.meas_sites = meas_sites
        
    def update_gate(self, gate_idx, new_array): 
        self.gate_list[gate_idx].matrix = new_array
        if self.layers is not None:
            self.construct_layers()
        
    def decompose(self, basis='ZSX'): 
        gate_list = []
        for gate in self.gate_list:
            if gate.name == "U" and gate.angles is not None:
                decomposed_gates = gate.decompose(basis)
                gate_list.extend(decomposed_gates)
            else:
                gate_list.append(gate)

        self.gate_list = gate_list
        
    def construct_layers(self): 
        layers = []
        for gate in self.gate_list:
            place_available = True
            layer_ind = len(layers) - 1

            # first we find the oldest layer that fails to commute with current gate
            while place_available and layer_ind >= 0:
                g_indices = set(gate.indices)
                existing_indices = {ind for gate in layers[layer_ind] for ind in gate.indices}
                if g_indices.intersection(existing_indices): 
                    place_available = False
                else: 
                    layer_ind -= 1

            # then we add current gate to the following layer
            if layer_ind < len(layers) - 1: 
                layers[layer_ind + 1].append(gate)
            else: 
                layers.append([gate])

        self.layers = layers
    
    def add_noise(self, nm): 
        layers = []
        for layer in self.layers: 
            if not any(isinstance(obj, SuperOp) for obj in layer):
                if isinstance(nm, ThermalNoise): 
                    layer_time = max([gate_time(gate, nm) for gate in layer])
                    active_qubits = set(np.concatenate([gate.indices for gate in layer]))
                    idle_qubits = set(range(self.num_sites)) - active_qubits
                    noise_layer = [ID(nm.local_dim, layer_time, i, nm=nm, dd=(i in idle_qubits))
                                    for i in range(self.num_sites)] 
                    
                else: 
                    noise_layer = [ID(2, 0.0, [i for i in range(self.num_sites)], nm=nm)] 
                
                layer.extend(noise_layer)
                layers.append(layer)
            else: 
                layers.append(layer)
        
        self.layers = layers
        
    def build(self, nm=None):
        """ this changes the current circuit instance and returns it """
        if not self.built:
            self.decompose()
            self.construct_layers()
            if nm is not None: 
                self.add_noise(nm)
            self.built = True
        return self
    
    def old_build(self, nm=None): 
        """ DEPRECATED: note that this returns a new circuit instance built with provided noise model
        and leaves the current circuit instance unchanged """
        new_circ = copy.deepcopy(self)
        new_circ.decompose()
        new_circ.construct_layers()
        if nm is not None: 
            new_circ.add_noise(nm)
        
        new_circ.built = True    
        return new_circ
    
    def add_meas(self, meas_sites): 
        self.meas_sites = meas_sites
        
    def dag(self): 
        gate_list = [g.dag() for g in self.gate_list[::-1]]
        return Circuit(gate_list, n=self.num_sites, local_dim=self.local_dim)
        
    def to_matrix(self): 
        return circ_to_mat(self)
        
    def __str__(self): 
        if self.layers is not None: 
            full_str = []
            for layer in self.layers:
                layer_str = " \n".join([str(gate) for gate in layer])
                full_str.append(layer_str)
            full_str = " \n\n".join(full_str)
            return full_str
        
        else: 
            return " \n".join([str(gate) for gate in self.gate_list])  

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.gate_list[key]
        elif isinstance(key, slice):
            return Circuit(self.gate_list[key], n=self.num_sites, local_dim=self.local_dim)
        else:
            raise TypeError("Invalid index type.")
        
    def __len__(self):
        return len(self.gate_list)
    
    def __add__(self, other):
        return compose_circuits([self, other])
        
    def draw(self): # TODO 
        """ use svgwrite to visualize the circuit """
        return None
    
    def copy(self):
        new_gate_list = [gate.shallow_copy() for gate in self.gate_list]
        new_circ = Circuit(new_gate_list, n=self.num_sites, local_dim=self.local_dim, 
                       meas_sites=self.meas_sites)
        if self.layers is not None:
            new_circ.layers = [[g.shallow_copy() for g in layer] for layer in self.layers]
        new_circ.built = self.built
        return new_circ

def compose_circuits(circ_list): 
    gate_list = [circ.gate_list for circ in circ_list]
    gate_list = [gate for sublist in gate_list for gate in sublist]
    return Circuit(gate_list, n=circ_list[0].num_sites)

def circ_to_mat(circ, n=None): 
    if circ.layers == None: 
        circ.construct_layers()
    n = circ.num_sites if n is None else n
    
    layer_mats = []    
    for layer in circ.layers: 
        missing_indices = set(range(n)) - set(np.concatenate([gate.indices for gate in layer]))
        for index in missing_indices: 
            layer.append(ID(circ.local_dim, 0.0, [index], None))
        
        sorted_layer = sorted(layer, key=lambda gate: gate.indices[0])
        mat_list = [gate.matrix for gate in sorted_layer]
        layer_mats.append(reduce(np.kron, mat_list))
        
    return np.linalg.multi_dot(layer_mats[::-1]) if len(layer_mats) > 1 else layer_mats[0]

def shift_gate_list(circ, start_site):
    """ return a new gate list with all gate indices shifted by start_site."""
    def _shifted_gate(gate, start_site):
        new_gate = copy.deepcopy(gate)
        new_gate.indices = np.array(gate.indices) + start_site
        return new_gate
    
    return [
        _shifted_gate(gate, start_site)
        for gate in circ.gate_list
    ]

def two_local_circ(skeleton, params=None, mag=None):
    gate_list = []
    for i,indices in enumerate(skeleton):
        if mag is None and params is None:
            mat = random_unitary(4)
        elif params is not None:
            mat = kak_unitary(params[15*i:15*(i+1)])
        else:
            mat = kak_unitary(np.random.rand(15) * mag)

        gate_list.append(Gate("rand_U", indices, mat))
    return Circuit(gate_list)